/* ========================================================================
 * Copyright (C) 2000-2004  Technische Universitaet Muenchen
 * License: GPL v2 (or later)
 *
 * For further information please contact the DWARF team at
 * <dwarf-gnu@augmentedreality.de>
 *
 * EkfWindowImpl.h
 * Daniel Pustka <pustka@in.tum.de> May 2004
 * Distributed Wearable Augmented Reality Framework - www.augmentedreality.de
 * For documentation and instructions, see the DWARF documentation website at
 * http://wwwbruegge.in.tum.de/projects/dwarf/doc/
 *
 * For further questions about DWARF, please contact us at
 * <dwarf@augmentedreality.de>
 * For questions about this particular service, contact
 * Daniel Pustka <pustka@in.tum.de>
 *
 * $Id: AutoTuner.cpp,v 1.2 2004/08/14 23:32:39 pustka Exp $
 */

#include "AutoTuner.h"
#include "PoseSequenceView.h"
#include <debug.h>
#include <TimeUtil.h>
#include <iostream>
#include <iomanip>

#ifdef HAVE_GSL
#include <gsl/gsl_multimin.h>
#endif 

using DWARF::Time;
using DWARF::PoseData;
using std::vector;
using std::cout;


AutoTuner::AutoTuner( const PoseSequenceSubCollection& Sources,
   const PoseEKF::MotionModel& MoMo, int nPredict, bool bPosXYZIndependent )
   : m_Sequences( Sources )
   , m_MoMo( MoMo)
   , m_nPredict( nPredict )
   , m_bPosXYZIndependent( bPosXYZIndependent )
   {
   }


double AutoTuner::ComputePredictionError( double& fAvgErr )
   {
   double fDiffSum = 0.0;
   double fSqDiffSum = 0.0;
   PoseEKF EKF( m_MoMo );

   // compute a predicted state every 10 ms
   Time PredictTime = { std::abs( m_nPredict ) / 1000, ( std::abs( m_nPredict ) % 1000 ) * 1000 };

   // vector with iterator for each sequence
   vector<PoseSequence::SequenceType::const_iterator> Its;

   // lock all sequences and push iterator into vector
   for ( unsigned i = 0; i < m_Sequences.size(); i++ )
      {
      m_Sequences[ i ]->GetMutex().lock();
      Its.push_back( m_Sequences[ i ]->begin() );
      }

   // find first event
   Time ItTime = Its[ 0 ]->timestamp;
   int iIt = 0;
   for ( unsigned i = 1; i < Its.size(); i++ )
      if ( Its[ i ]->timestamp < ItTime )
         ItTime = Its[ i ]->timestamp, iIt = i;

   Time NextPredict = Its[ iIt ]->timestamp;

   int nMeasurements = -10;
   while ( Its[ 0 ] != m_Sequences[ 0 ]->end() && iIt >= 0)
      {
      // measure
      EKF.AddPose( *Its[ iIt ] );

      // predict only after ekf has stabilized
      if ( ++nMeasurements > 0 )
         {
         // compute predicted time
         Time FutureTime;
         PoseSequence::SequenceType::const_iterator it1 = Its[ 0 ], it2 = Its[ 0 ];

         if ( m_nPredict < 0 )
            {
            FutureTime = ItTime - PredictTime;

            // search for the two measurements around the predicted time
            it1--;
            while ( it1 != m_Sequences[ 0 ]->begin() && FutureTime < it1->timestamp )
               it1--, it2--;

            if ( it1 == m_Sequences[ 0 ]->begin() )
               break;
            }
         else
            {
            FutureTime = ItTime + PredictTime;

            // search for the two measurements around the predicted time
            it2++;
            while ( it2 != m_Sequences[ 0 ]->end() && it2->timestamp < FutureTime )
               it1++, it2++;

            if ( it2 == m_Sequences[ 0 ]->end() )
               break;
            }

         // predict
         PoseData Predicted = EKF.PredictPose( FutureTime );

         // compute time differences
         Time TDiff1 = FutureTime - it1->timestamp;
         double fT1 = TDiff1.seconds + TDiff1.microseconds / 1000000.0;
         Time TDiff2 = it2->timestamp - it1->timestamp;
         double fT2 = TDiff2.seconds + TDiff2.microseconds / 1000000.0;

         // weights
         double fW2 = fT1 / fT2;
         double fW1 = 1.0 - fW2;

         // compute linearly interpolated position
         if ( m_MoMo.GetPosDerivations() >= 0 )
            {
            double fDiffSq = 0;
            for ( int i = 0; i < 3; i++ )
               {
               // interpolated and compute difference to predicted
               double fDiff = it1->position[ i ] * fW1 + it2->position[ i ] * fW2;
               fDiff -= Predicted.position[ i ];
               fDiffSq += fDiff * fDiff;
               }

            // add to squared difference
            fSqDiffSum += fDiffSq;
            fDiffSum += sqrt( fDiffSq );
            }

         // compute SLERP interpolated orientation
         if ( m_MoMo.GetOriDerivations() >= 0 )
            {
            double QInterp[ 4 ];
            DWARF::Util::quaternionSlerp( QInterp, it1->orientation, it2->orientation, fW2 );
            DWARF::Util::quaternionNegateIfCloser( QInterp, Predicted.orientation );

            // compute difference
            double fDiffSq = 0.0;
            for ( int i = 0; i < 4; i++ )
               {
               double d = QInterp[ i ] - Predicted.orientation[ i ];
               fDiffSq += d * d;
               }

            // add to squared difference
            fSqDiffSum += fDiffSq;
            fDiffSum += sqrt( fDiffSq );
            }
         }

      // advance file
      Its[ iIt ]++;

      // find next measurement
      ItTime.seconds = 0x7FFFFFFF;
      iIt = -1;
      for ( unsigned i = 0; i < Its.size(); i++ )
         if ( Its[ i ] != m_Sequences[ i ]->end() && Its[ i ]->timestamp < ItTime )
            ItTime = Its[ i ]->timestamp, iIt = i;
      }

   // unlock sequences
   for ( unsigned i = 0; i < m_Sequences.size(); i++ )
      m_Sequences[ i ]->GetMutex().unlock();

   fAvgErr = fDiffSum / nMeasurements;
   return fSqDiffSum / nMeasurements;
   }

#ifdef HAVE_GSL

double AutoTuner::my_opt_f( const gsl_vector* v, void* par )
   {
   AutoTuner* pPar = reinterpret_cast<AutoTuner*>( par );
   const int nPosValues = ( pPar->m_MoMo.GetPosDerivations() + 1 ) * ( pPar->m_bPosXYZIndependent ? 3 : 1 );

   // copy to process noise
   if ( pPar->m_bPosXYZIndependent )
      for ( int i = 0; i < nPosValues; i++ )
         pPar->m_MoMo.SetPosProcessNoise( i % 3, i / 3, gsl_vector_get( v, i ) );
   else
      for ( int i = 0; i < nPosValues; i++ )
         for ( int j = 0; j < 3; j++ )
            pPar->m_MoMo.SetPosProcessNoise( j, i, gsl_vector_get( v, i ) );

   for ( int i = 0; i < pPar->m_MoMo.GetOriDerivations() + 1; i++ )
      pPar->m_MoMo.SetOriProcessNoise( i, gsl_vector_get( v, nPosValues + i ) );

   double fDummy;
   return pPar->ComputePredictionError( fDummy );
   }


void AutoTuner::OptimizePN()
   {
   const int nPosValues = ( m_MoMo.GetPosDerivations() + 1 ) * ( m_bPosXYZIndependent ? 3 : 1 );
   const int np = nPosValues + m_MoMo.GetOriDerivations() + 1;

   // variables
   const gsl_multimin_fminimizer_type *T = gsl_multimin_fminimizer_nmsimplex;

   // starting point
   gsl_vector* x = gsl_vector_alloc( np );
   if ( m_bPosXYZIndependent )
      for ( int i = 0; i < nPosValues; i++ )
         gsl_vector_set( x, i, m_MoMo.GetPosProcessNoise( i % 3, i / 3 ) );
   else
      for ( int i = 0; i < nPosValues; i++ )
         gsl_vector_set( x, i, m_MoMo.GetPosProcessNoise( 0, i ) );

   for ( int i = 0; i < m_MoMo.GetOriDerivations() + 1; i++ )
      gsl_vector_set( x, i + nPosValues, m_MoMo.GetOriProcessNoise( i ) );

   // set step size
   gsl_vector* ss = gsl_vector_alloc( np );
   for ( int i = 0; i < np; i++ )
      gsl_vector_set( ss, i, gsl_vector_get( x, i ) / 2.3984 );

   // initialize method and iterate
   gsl_multimin_function minex_func;
   minex_func.f = my_opt_f;
   minex_func.n = np;
   minex_func.params = this;

   gsl_multimin_fminimizer *s = gsl_multimin_fminimizer_alloc( T, np );
   gsl_multimin_fminimizer_set( s, &minex_func, x, ss );

   size_t iter = 0;
   int status;
   do
      {
      iter++;
      status = gsl_multimin_fminimizer_iterate( s );

      if ( status )
         break;

      // find minimal value
      double fMin = fabs( gsl_vector_get( s->x, 0 ) );
      for ( int i = 1; i < np; i++ )
         fMin = std::min( fMin, fabs( gsl_vector_get( s->x, i ) ) );

      double size = gsl_multimin_fminimizer_size( s );
      status = gsl_multimin_test_size( size, fMin / 10 );

      if ( status == GSL_SUCCESS )
         cout << "Converged to minimum at" << std::endl;

      cout << iter << ": ";
      for ( int i = 0; i < np; i++ )
         cout << std::setw( 6 ) << std::setfill( '0' ) << gsl_vector_get( s->x, i ) << " ";
      cout << "f = " << s->fval << " size = " << size << std::endl;
      }
      while ( status == GSL_CONTINUE && iter < 300 );

   // copy to process noise
   if ( m_bPosXYZIndependent )
      for ( int i = 0; i < nPosValues; i++ )
         m_MoMo.SetPosProcessNoise( i % 3, i / 3, fabs( gsl_vector_get( s->x, i ) ) );
   else
      for ( int i = 0; i < nPosValues; i++ )
         for ( int j = 0; j < 3; j++ )
            m_MoMo.SetPosProcessNoise( j, i, fabs( gsl_vector_get( s->x, i ) ) );

   for ( int i = 0; i < m_MoMo.GetOriDerivations() + 1; i++ )
      m_MoMo.SetOriProcessNoise( i, fabs( gsl_vector_get( s->x, nPosValues + i ) ) );

   // cleanup
   gsl_vector_free( x );
   gsl_vector_free( ss );
   gsl_multimin_fminimizer_free( s );
   }

#endif
