/***************************************************************************
 *                         Ekf.cpp  -  description
 *                            -------------------
 *   begin                : 2004/05/22
 *   copyright            : (C) 2004 by Daniel Pustka
 *   email                : pustka@in.tum.de
 *
 *   $Id: Ekf.cpp,v 1.1 2004/06/07 13:34:29 pustka Exp $
 *   $Revision: 1.1 $
 *
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <debug.h>
#include "Ekf.h"

namespace JAMA_Sucks {
// jama_lu.h includes using namespace TNT
#include <jama_lu.h>
}
namespace JAMA {
   using namespace JAMA_Sucks::JAMA;
};

using TNT::matmult;

typedef TNT::Array2D<double> Matrix;
typedef TNT::Array1D<double> Vector;

// TNT is crap. There is no matrix transpose.
static Matrix transpose( const Matrix& M )
	{
	int n = M.dim1();
	int m = M.dim2();

	Matrix Mt( m, n );

	for ( int i = 0; i < n; i++ )
		for ( int j = 0; j < m; j++ )
			Mt[j][i] = M[i][j];

	return Mt;
	}



ExtendedKalmanFilter::ExtendedKalmanFilter( int nStateSize )
   : m_nStateSize( nStateSize )
   , m_State( nStateSize )
   , m_Covariance( nStateSize, nStateSize )
   , m_ProcessNoise( nStateSize )
   {
   }


void ExtendedKalmanFilter::AddTimeUpdateFunction( std::pair<int, int> OutRange, std::pair<int, int> InRange,
   std::auto_ptr<TimeUpdateFunction> pTUF )
   {
   m_TUFs.push_back( pTUF );
   m_TUFInfos.push_back( TUFInfo( OutRange, InRange ) );
   }


void ExtendedKalmanFilter::SetProcessNoise( const double* pNoise )
   {
   m_ProcessNoise.inject( Vector( m_nStateSize, const_cast<double*>( pNoise ) ) );
   DEBUGSTREAM( 40, "SetProcessNoise P = " << m_ProcessNoise );
   }


void ExtendedKalmanFilter::Initialize( const double* pValues, const double* pSigma )
   {
   m_State.inject( Vector( m_nStateSize, const_cast<double*>( pValues ) ) );
   m_Covariance = 0.0;
   for ( int i = 0; i < m_nStateSize; i++ )
      m_Covariance[ i ][ i ] = pSigma[ i ] * pSigma[ i ];
   }


void ExtendedKalmanFilter::TimeUpdate( double fTimeDelta )
   {
   DEBUGSTREAM( 25, "EKF TimeUpdate: deltaTime = " << fTimeDelta );

   // initialize new state vector
   Vector NewState( m_nStateSize );

   // initialize Jacobian
   Matrix J( m_nStateSize, m_nStateSize );
   J = 0.0;

   for ( unsigned i = 0; i < m_TUFs.size(); i++ )
      {
      // apply all time update functions
      int iOutStart = m_TUFInfos[ i ].OutRange.first;
      int iInStart = m_TUFInfos[ i ].InRange.first;
      Matrix A = m_TUFs[ i ]->Predict( fTimeDelta, &m_State[ iInStart ], &NewState[ iOutStart ] );

      // TNT is crap: No need to copy if submatrix() would work!
      for ( int j = 0; j < m_TUFInfos[ i ].OutRange.second; j++ )
         for ( int k = 0; k < m_TUFInfos[ i ].InRange.second; k++ )
            J[ j + iOutStart ][ k + iInStart ] = A[ j ][ k ];
      }

   DEBUGSTREAM( 25, "EKF TimeUpdate: x^ = " << NewState );

   // compute new covariance
   DEBUGSTREAM( 30, "TimeUpdate Jacobi = " << J );
   Matrix NewCov = matmult( matmult( J, m_Covariance ), transpose( J ) );

   // add process noise
   for ( int i = 0; i < m_nStateSize; i++ )
      NewCov[ i ][ i ] += m_ProcessNoise[ i ] * m_ProcessNoise[ i ] * fTimeDelta * fTimeDelta;

   DEBUGSTREAM( 27, "EKF TimeUpdate: P^ = " << NewCov );

   // copy state and covariance
   m_State = NewState;
   m_Covariance = NewCov;
   }


TNT::Array2D<double> ExtendedKalmanFilter::Predict( double fTimeDelta,
   const std::list<std::pair<int, int> >& Range, double* pOut )
   {
   // compute output state vector size
   int nOutputSize = 0;
   for ( std::list<std::pair<int, int> >::const_iterator it = Range.begin(); it != Range.end(); it++ )
      nOutputSize += it->second;

   // initialize Jacobian
   Matrix J( nOutputSize, m_nStateSize );
   J = 0.0;

   int iOut = 0;
   for ( std::list<std::pair<int, int> >::const_iterator it = Range.begin(); it != Range.end(); it++ )
      {
      for ( unsigned i = 0; i < m_TUFs.size(); i++ )
         {
         int iOutStart = m_TUFInfos[ i ].OutRange.first;

         // check if TUF overlaps with requested region
         if ( iOutStart >= it->first &&
            iOutStart + m_TUFInfos[ i ].OutRange.second <= it->first + it->second )
            {
            int iInStart = m_TUFInfos[ i ].InRange.first;
            int iDest = iOut + iOutStart - it->first;

            // apply all time update functions
            Matrix A = m_TUFs[ i ]->Predict( fTimeDelta, &m_State[ iInStart ], pOut + iDest);

            // TNT is crap: No need to copy if submatrix() would work!
            for ( int j = 0; j < m_TUFInfos[ i ].OutRange.second; j++ )
               for ( int k = 0; k < m_TUFInfos[ i ].InRange.second; k++ )
                  J[ j + iDest ][ k + iInStart ] = A[ j ][ k ];
            }
         }

      iOut += it->second;
      }

   DEBUGSTREAM( 30, "Prediction Jacobi = " << J );

   // compute new covariance
   Matrix NewCov = matmult( matmult( J, m_Covariance ), transpose( J ) );

   // add process noise
   iOut = 0;
   for ( std::list<std::pair<int, int> >::const_iterator it = Range.begin(); it != Range.end(); it++ )
      for ( int i = 0; i < it->second; i++, iOut++ )
         NewCov[ iOut ][ iOut ] += m_ProcessNoise[ it->first + i ] * m_ProcessNoise[ it->first + i ] *
            fTimeDelta * fTimeDelta;

   return NewCov;
   }


void ExtendedKalmanFilter::Measure( int nMeasurementSize, std::pair<int, int> InRange,
   const MeasurementFunction& MF, const double* pMeasurement, const TNT::Array2D<double>& Covariance )
   {
   const Matrix Measurement( nMeasurementSize, 1, const_cast<double*>( pMeasurement ) );

   DEBUGSTREAM( 25, "EKF Measure: z = " << Measurement );

   // predict measurement
   Matrix Diff( nMeasurementSize, 1 );
   Matrix H = MF.PredictMeasurement( &m_State[ InRange.first ], Diff[ 0 ] );

   // TNT is crap: unnecessary copying because submatrix() is defunct!
   if ( H.dim2() != m_nStateSize )
      {
      Matrix H2( nMeasurementSize, m_nStateSize );
      H2 = 0.0;
      for ( int i = 0; i < nMeasurementSize; i++ )
         for ( int j = 0; j < H.dim2(); j++ )
            H2[ i ][ j + InRange.first ] = H[ i ][ j ];
      H = H2;
      }

   DEBUGSTREAM( 30, "H = " << H );

   // compute kalman gain
   Matrix Ht = transpose( H );
   Matrix HP = matmult( H, m_Covariance );
   Matrix S = Covariance + matmult( HP, Ht );

   DEBUGSTREAM( 30, "S = " << S );

   // Note: if we were REALLY sure that S is symmetric, the cholesky decomposition
   // would be a better choice
   JAMA::LU<double> Lu( S );
   Matrix K = transpose( Lu.solve( HP ) );

   DEBUGSTREAM( 30, "K = " << K );

   if ( !Lu.isNonsingular() )
      {
      std::string s( "Singular matrix - cannot compute kalman gain!" );
      DEBUGSTREAM( 1, s );
      throw( s );
      }

   // compute measurement difference
   Diff = Measurement - Diff;

   // compute new state
   Matrix StateAdd = matmult( K, Diff );
   m_State += Vector( m_nStateSize, StateAdd[ 0 ] );

   // update covariance
   m_Covariance -= matmult( K, HP );

   DEBUGSTREAM( 25, "EKF Measure: x = " << m_State );
   DEBUGSTREAM( 27, "EKF Measure: P = " << m_Covariance );
   }
