/***************************************************************************
 *                         Ekf.h  -  description
 *                            -------------------
 *   begin                : 2004/05/21
 *   copyright            : (C) 2004 by Daniel Pustka
 *   email                : pustka@in.tum.de
 *
 *   $Id: Ekf.h,v 1.1 2004/06/07 13:34:29 pustka Exp $
 *   $Revision: 1.1 $
 *
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef __EKF_H__
#define __EKF_H__

// autoconf generated file with configuration information
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

// TNT is crap. Can't link more than one file that includes it.
#define STOPWATCH_H
#include <tnt.h>

#include <list>
#include <memory>
#include "auto_vector.h"


/**
 * Implements a generic extended kalman filter
 * Time update and measurement functions must be user-supplied by implementing
 * the TimeUpdateFunction and MeasurementFunction interfaces.
 * It is possible to supply different TimeUpdateFunctions for different
 * subvectors of the state vector to allow a modular construction of EKFs.
 */
class ExtendedKalmanFilter
   {
   public:
      /**
       * An interface for time update functions (TUFs).
       * Each TUF takes a range of the EKF's state variables and returns a possibly
       * different range of updated states. The ranges are specified when calling
       * ExtendedKalmanFilter::AddTimeUpdateFunction.
       */
      class TimeUpdateFunction
         {
         public:
            /**
             * Predicts state pOut from state pIn
             * @param fTimeDelta The delta time for the prediction
             * @param pIn Points to the first input variable
             * @param pOut Points to the first output variable
             * @return The Jacobian of the prediction evaluated at the current state.
             *    An n*m matrix where n is the number of output variables and m the
             *    number of inputs.
             */
            virtual TNT::Array2D<double>
               Predict( double fTimeDelta, const double* pIn, double* pOut ) = 0;

            /** just to make sure it's virtual */
            virtual ~TimeUpdateFunction() {}
         };

      /**
       * An interface for measurement functions (MFs)
       */
      class MeasurementFunction
         {
         public:
            /**
             * Computes the predicted measurement pOut from the predicted state pIn
             * @param pIn Pointer to the first input variable
             * @param pOut Pointer to the predicted measurement
             * @return The Jacobian of the MF evaluated at the state pIn
             */
            virtual TNT::Array2D<double>
               PredictMeasurement( const double* pIn, double* pOut ) const = 0;

            /** just to make sure it's virtual */
            virtual ~MeasurementFunction() {}
         };

   protected:
      // general info
      int m_nStateSize;

      // EKF state
      TNT::Array1D<double> m_State;
      TNT::Array2D<double> m_Covariance;

      // process noise
      TNT::Array1D<double> m_ProcessNoise;

      // TimeUpdateFunctions
      struct TUFInfo
         {
         std::pair<int, int> OutRange;
         std::pair<int, int> InRange;

         TUFInfo( std::pair<int, int> _OR, std::pair<int, int> _IR )
            { OutRange = _OR; InRange = _IR; }
         };

      auto_vector<TimeUpdateFunction> m_TUFs;
      std::vector<TUFInfo> m_TUFInfos;

   public:

      /**
       * Initializes the EKF
       * @param nStateSize Number of state variables
       */
      ExtendedKalmanFilter( int nStateSize );

      /**
       * Adds a time update function to the EKF
       * To enable the EKF to be constructed dynamically out of predefined blocks
       * we allow different TUFs to be specified for different state variables.
       * @param OutRange state variables updated by the TUF: <first, count>
       * @param InRange state variables needed by the TUF: <first, count>
       * @param pTUF Pointer to TUF - EKF takes ownership of the object
       */
      void AddTimeUpdateFunction( std::pair<int, int> OutRange, std::pair<int, int> InRange,
         std::auto_ptr<TimeUpdateFunction> pTUF );

      /**
       * Sets the process noise
       * @param pNoise Array with an entry for each state variable containing the standard
       *    deviation per time unit.
       */
      void SetProcessNoise( const double* pNoise );

      /**
       * Initializes the EKF
       * @param pValues Initial values of the state vector entries
       * @param pSigma Standard deviations of the state vector entries
       */
      void Initialize( const double* pValues, const double* pSigma );

      /**
       * Advances the state by fTimeDelta
       * @param fTimeDelta specifies the time difference
       */
      void TimeUpdate( double fTimeDelta );

      /**
       * Returns a predicted state sub-vector without updating the internal state
       * Note: In the current implementation, requested subvectors must cover entire
       * time update functions. It is not possible to request only a part of one TUF.
       * @param fTimeDelta how much to predict into the future
       * @param Range subset of state vector to be returned. pair is <firstIndex, size>
       * @param pOut Location where to write
       * @return Covariance of predicted state
       */
      TNT::Array2D<double> Predict( double fTimeDelta, const std::list< std::pair<int, int> >& Range,
         double* pOut );

      /**
       * Performs a measurement update
       * @param nMeasurementSize Number of variables in the measurement
       * @param InRange Range of state variables required for the measurement
       * @param MF the measurement function
       * @param pMeasurement Points to the measurement data. Size is determined by the size of
       *    the Jacobian returned by the MF
       * @param Covariance The covariance of the measurement data
       */
      void Measure( int nMeasurementSize, std::pair<int, int> InRange, const MeasurementFunction& MF,
         const double* pMeasurement, const TNT::Array2D<double>& Covariance );

      /**
       * returns the state vector
       */
      const TNT::Array1D<double>& GetState() const
         { return m_State; }

      /**
       * returns the covariance matrix
       */
      const TNT::Array2D<double>& GetCovariance() const
         { return m_Covariance; }

      /**
       * returns the state vector size
       */
      int GetStateSize() const
         { return m_nStateSize; }
   };

#endif
