/***************************************************************************
 *                         EkfMeasure.cpp  -  description
 *                            -------------------
 *   begin                : 2004/05/22
 *   copyright            : (C) 2004 by Daniel Pustka
 *   email                : pustka@in.tum.de
 *
 *   $Id: EkfMeasure.cpp,v 1.1 2004/06/07 13:34:29 pustka Exp $
 *   $Revision: 1.1 $
 *
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "EkfMeasure.h"
#include <math.h>
#include <algorithm>

typedef TNT::Array2D<double> Matrix;


////////////////////////////////////////////////////////////////////////////
// IdentityMeasurement

IdentityMeasurement::IdentityMeasurement( int nVariables )
   : m_nVariables( nVariables )
   {
   }


TNT::Array2D<double> IdentityMeasurement::PredictMeasurement( const double* pIn, double* pOut ) const
   {
   Matrix J( m_nVariables, m_nVariables );
   J = 0.0;

   for ( int i = 0; i < m_nVariables; i++ )
      {
      pOut[ i ] = pIn[ i ];
      J[ i ][ i ] = 1.0;
      }

   return J;
   }


////////////////////////////////////////////////////////////////////////////
// QuaternionMeasurement

TNT::Array2D<double> QuaternionMeasurement::PredictMeasurement( const double* pIn, double* pOut ) const
   {
   // normalize
   double fD = 0.0;
   for ( int i = 0; i < 4; i++ )
      fD += pIn[ i ] * pIn[ i ];

   double fL = sqrt( fD );

   for ( int i = 0; i < 4; i++ )
      pOut[ i ] = pIn[ i ] / fL;

   // compute jacobian
   // see Azuma's diss, page 98
   Matrix J( 4, 4 );
   for ( int i = 0; i < 4; i++ )
      for ( int j = 0; j < 4; j++ )
         J[ i ][ j ] = - pIn[ i ] * pIn[ j ] / ( fL * fD );

   for ( int i = 0; i < 4; i++ )
      J[ i ][ i ] += fL / fD;

   return J;
   }


////////////////////////////////////////////////////////////////////////////
// CombinedMeasurement

CombinedMeasurement::CombinedMeasurement(
   int nMeasurementSize1, std::pair<int, int> InRange1,
   std::auto_ptr<ExtendedKalmanFilter::MeasurementFunction> pMF1,
   int nMeasurementSize2, std::pair<int, int> InRange2,
   std::auto_ptr<ExtendedKalmanFilter::MeasurementFunction> pMF2 )
   {
   m_MFInfos.push_back( MFInfo( nMeasurementSize1, InRange1 ) );
   m_MFs.push_back( pMF1 );
   m_MFInfos.push_back( MFInfo( nMeasurementSize2, InRange2 ) );
   m_MFs.push_back( pMF2 );
   }


TNT::Array2D<double> CombinedMeasurement::PredictMeasurement( const double* pIn, double* pOut ) const
   {
   // compute size of measurement and jacobian
   int nSize = 0;
   int nEnd = 0;
   for ( unsigned i = 0; i < m_MFInfos.size(); i++ )
      {
      nSize += m_MFInfos[ i ].nMeasurementSize;
      nEnd = std::max( nEnd, m_MFInfos[ i ].InRange.first + m_MFInfos[ i ].InRange.second );
      }

   // execute all measurement functions
   Matrix J( nSize, nEnd );
   J = 0.0;
   int iMeasurement = 0;
   for ( unsigned i = 0; i < m_MFs.size(); i++ )
      {
      Matrix Jp = m_MFs[ i ]->PredictMeasurement(
         pIn + m_MFInfos[ i ].InRange.first, pOut + iMeasurement );

      // TNT is crap: submatrix!
      // copy partial jacobians
      for ( int k = 0; k < Jp.dim1(); k++ )
         for ( int j = 0; j < Jp.dim2(); j++ )
            J[ iMeasurement + k ][ m_MFInfos[ i ].InRange.first + j ] = Jp[ k ][ j ];

      iMeasurement += m_MFInfos[ i ].nMeasurementSize;
      }

   return J;
   }
