/***************************************************************************
 *                         EkfPredict.cpp  -  description
 *                            -------------------
 *   begin                : 2004/05/22
 *   copyright            : (C) 2004 by Daniel Pustka
 *   email                : pustka@in.tum.de
 *
 *   $Id: EkfPredict.cpp,v 1.1 2004/06/07 13:34:29 pustka Exp $
 *   $Revision: 1.1 $
 *
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <debug.h>
#include "EkfPredict.h"
#include <string.h>
#include <dwarfutil.h>

typedef TNT::Array2D<double> Matrix;


////////////////////////////////////////////////////////////////////////////
// PolynomialPredictor

PolynomialPredictor::PolynomialPredictor( int nVariables, int nDerivations )
   : m_nVariables( nVariables )
   , m_nDerivations( nDerivations )
   {
   }


TNT::Array2D<double> PolynomialPredictor::Predict( double fTimeDelta, const double* pIn, double* pOut )
   {
   Matrix J( m_nVariables, m_nVariables * ( m_nDerivations + 1 ) );
   J = 0.0;

   for ( int i = 0; i < m_nVariables; i++ )
      {
      double f = 1.0;
      double x = 0;
      int k = i;

      for ( int j = 0; j <= m_nDerivations; j++ )
         {
         x += pIn[ k ] * f;
         J[ i ][ k ] = f;
         f *= fTimeDelta / ( j + 1 );
         k += m_nVariables;
         }

      pOut[ i ] = x;
      }

   DEBUGSTREAM( 50, "PolyPredict J = " << J );

   return J;
   }


////////////////////////////////////////////////////////////////////////////
// QuaternionPredictor

QuaternionPredictor::QuaternionPredictor( int nDerivations )
   : m_nDerivations( nDerivations )
   {
   }


TNT::Array2D<double> QuaternionPredictor::Predict( double fTimeDelta, const double* pIn, double* pOut )
   {
   Matrix J( 4, 4 + 3 * m_nDerivations );

   // compute logarithm of difference quaternion
   double DiffQLog[ 3 ];
   double fOmega = 0.0;

   for ( int i = 0; i < 3; i++ )
      {
      double f = fTimeDelta;
      double x = 0.0;
      int k = 4 + i;

      for ( int j = 0; j < m_nDerivations; j++ )
         {
         x += pIn[ k ] * f;
         f *= fTimeDelta / ( j + 2 );
         k += 3;
         }

      DiffQLog[ i ] = x;
      fOmega += x * x;
      }

   DEBUGSTREAM( 50, "QuaternionPredict: DiffQLog = ["<<DiffQLog[0]<<","<<DiffQLog[1]<<","<<DiffQLog[2]<<"]" );

   // fOmega = abs( DiffQLog )
   fOmega = sqrt( fOmega );
   DEBUGSTREAM( 60, "QuaternionPredict: fOmega = " << fOmega );

   // useful variables
   double fSinc1_2Omega = fOmega < 1e-6 ? 0.5 : sin( fOmega / 2 ) / fOmega;
   double fCos1_2Omega = cos( fOmega / 2 );

   DEBUGSTREAM( 60, "QuaternionPredict: sinc(omega/2) = " << fSinc1_2Omega );
   DEBUGSTREAM( 60, "QuaternionPredict: cos(omega/2) = " << fCos1_2Omega );

   // compute difference quaternion
   double DiffQ[ 4 ];
   DiffQ[ 0 ] = DiffQLog[ 0 ] * fSinc1_2Omega;
   DiffQ[ 1 ] = DiffQLog[ 1 ] * fSinc1_2Omega;
   DiffQ[ 2 ] = DiffQLog[ 2 ] * fSinc1_2Omega;
   DiffQ[ 3 ] = fCos1_2Omega;

   // multiply to old quaternion
   DWARF::Util::multiplyQuaternion( pOut, pIn, DiffQ );

   // compute jacobian
   J[ 0 ][ 0 ] = fCos1_2Omega;
   J[ 0 ][ 1 ] =   DiffQLog[ 2 ] * fSinc1_2Omega;
   J[ 0 ][ 2 ] = - DiffQLog[ 1 ] * fSinc1_2Omega;
   J[ 0 ][ 3 ] =   DiffQLog[ 0 ] * fSinc1_2Omega;

   J[ 1 ][ 0 ] = - DiffQLog[ 2 ] * fSinc1_2Omega;
   J[ 1 ][ 1 ] = fCos1_2Omega;
   J[ 1 ][ 2 ] =   DiffQLog[ 0 ] * fSinc1_2Omega;
   J[ 1 ][ 3 ] =   DiffQLog[ 1 ] * fSinc1_2Omega;

   J[ 2 ][ 0 ] =   DiffQLog[ 1 ] * fSinc1_2Omega;
   J[ 2 ][ 1 ] = - DiffQLog[ 0 ] * fSinc1_2Omega;
   J[ 2 ][ 2 ] = fCos1_2Omega;
   J[ 2 ][ 3 ] =   DiffQLog[ 2 ] * fSinc1_2Omega;

   J[ 3 ][ 0 ] = - DiffQLog[ 0 ] * fSinc1_2Omega;
   J[ 3 ][ 1 ] = - DiffQLog[ 1 ] * fSinc1_2Omega;
   J[ 3 ][ 2 ] = - DiffQLog[ 2 ] * fSinc1_2Omega;
   J[ 3 ][ 3 ] = fCos1_2Omega;

   // how the derivations relate to the measurement
   double f = fTimeDelta * fSinc1_2Omega;
   for ( int i = 0; i < m_nDerivations; i++ )
      {
      int j = 4 + 3 * i;
      J[ 0 ][ j     ] =   f * pIn[ 3 ];
      J[ 0 ][ j + 1 ] = - f * pIn[ 2 ];
      J[ 0 ][ j + 2 ] =   f * pIn[ 1 ];

      J[ 1 ][ j     ] =   f * pIn[ 2 ];
      J[ 1 ][ j + 1 ] =   f * pIn[ 3 ];
      J[ 1 ][ j + 2 ] = - f * pIn[ 0 ];

      J[ 2 ][ j     ] = - f * pIn[ 1 ];
      J[ 2 ][ j + 1 ] =   f * pIn[ 0 ];
      J[ 2 ][ j + 2 ] =   f * pIn[ 3 ];

      J[ 3 ][ j     ] = - f * pIn[ 0 ];
      J[ 3 ][ j + 1 ] = - f * pIn[ 1 ];
      J[ 3 ][ j + 2 ] = - f * pIn[ 2 ];

      f *= fTimeDelta / ( i + 2 );
      }

   DEBUGSTREAM( 50, "QuaternionPredict: J = " << J );

   return J;
   }
