/** this is a manual conversion of the DWARF file PoseData.idl to a cpp header */

/***************************************************************************
 *                         PoseData.idl  -  description
 *                            -------------------
 *   begin                : Mon Jul  8 09:59:39 MEST 2002
 *   copyright            : (C) 2002 by Martin Wagner
 *   email                : wagnerm@in.tum.de
 *
 *   $Id: PoseData.idl,v 1.15 2004/09/03 18:14:38 toennis Exp $
 *   $Revision: 1.15 $
 *
 *   This idl contains the PoseData data types needed by most DWARF 
 *   components that exchange location information.
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   NOTE: This file contains the new Ubitrack data types and interfaces   *
 *   This will supersede the current PoseData in the near future           *
 *                                                                         *
 *   Franz Strasser <strassef@in.tum.de>                                   *
 *                                                                         *
 ***************************************************************************/



/**
 * @defgroup Ubitrack Ubiquitous Tracking Environment
 * @brief All components for the Ubiquitous Tracking Environment (Ubitrack).
 *
 * The Ubitrack framework provides services and interfaces for setting
 * up a Ubiquitous Tracking Environment. Ubitrack serves as a framework for
 * Augmented Reality applications to provide a unique API for different
 * tracking technologies.
 * See the project's webpage for more details:
 * http://wwwbruegge.in.tum.de/DWARF/ProjectUbitrack
 *
 * @author Franz Strasser <strassef@in.tum.de>
 */


/**
 * @file PoseData.idl
 * @brief The IDL definition file for tracking data.
 *
 * This file contains all CORBA datatype definitions for the tracking
 * subsystem in DWARF. Currently, PoseData is used as standard format
 * for spatial information in all DWARF applications.
 * 
 * Additionally you can find here the newer datatypes for the Ubitrack
 * framework. These types will supersede the current PoseData once
 * Ubitrack is completely implemented in DWARF. Basically this file
 * contains all interfaces and datatypes as discussed on the
 * Sensor-API page:
 * http://wwwbruegge.in.tum.de/DWARF/SensorAPI
 *
 * @author Martin Wagner <wagnerm@in.tum.de>
 * @author Franz Strasser <strassef@in.tum.de>
 * @ingroup trackingidl
 * @ingroup Ubitrack
 */



#ifndef __POSEDATA_IDL
#define __POSEDATA_IDL

#include <DWARF/DwarfCommon.h>
#include <string>

namespace DWARF {

  /*****************
    Ubitrack types
  *****************/

  /**
   * @brief Data type for object identifiers in the Spatial
   * Relationship graph.
   *
   * An object identifier is a string which must be unique in a
   * pre-defined location context. In the first step, all object IDs
   * must be unique because we have only a global context.
   *
   * @ingroup Ubitrack
   */
  typedef std::string UTObjectID;

  /** Data type for 3D positions */
  typedef double pos3d[3];

  /** Data type for quaternions: x y z w */
  typedef double quaternion[4];

  /**
   * @brief Data type for 6x6 covariance matrices
   *
   * Covariance matrices represent statistical errors of spatial measurements in
   * one single matrix.
   *
   * @ingroup Ubitrack
   */
  typedef double covarMatrix6x6[6][6];

  /**
   * @brief The structure for the position data
   *
   * This structure defines the data which is delivered between the
   * tracking components. So we are talking about the Sensor-API
   * level. This is mainly discussed at
   * http://wwwbruegge.in.tum.de/projects/lehrstuhl/twiki/bin/view/DWARF/SensorAPI
   *
   * However as first step this data is also delivered to the
   * application, which is the Query-API level. This is not neccessary
   * the case in future implementations.
   *
   * @ingroup Ubitrack
   */
  struct PoseData {
    /** is the position field valid? */
    bool hasPosition;

    /** is the orientation field valid? */
    bool hasOrientation;

    /** The object ID of the source object */
    UTObjectID source;
    
    /** The object ID of the target object */
    UTObjectID target;

    /**
     * @brief Position of the target object
     *
     * Position of the target object in 3D space, based on the source
     * object's coordinate system.
     *
     */
    pos3d position;

    /**
     * @brief Orientation of the target object
     *
     * Orientation of the target object in 3D space, based on the
     * source object's coordinate system. Given as a quaternion with
     * the scalar of the quaternion being stored in orientation[3]
     *
     */
    quaternion orientation;

    /**
     * @brief Timestamp of the measurement
     *
     * The time the position expressed is valid, a pair of seconds and
     * nanoseconds, based on the standard UNIX time format.
     *
     */
    Time timestamp;

    /**
     * @brief Confidence value of the measurement
     *
     * A statistical value giving the probability that this
     * measurement exists.
     */
    double confidence;

    /**
     * @brief The covariance matrix of the measurement
     *
     * A 6x6 covariance matrix parameterizing the Gaussian error
     * distribution of the measured position and orientation.
     * The hasPosition/hasOrientation fields define which entries
     * are valid.
     */
    covarMatrix6x6 covariance;

    /** Standard deviation of measurement time */
    double timeError;
  };


};

#endif // __POSEDATA_IDL
