/* ========================================================================
 * Copyright (C) 2000-2004  Technische Universitaet Muenchen
 * License: GPL v2 (or later)
 *
 * For further information please contact the DWARF team at
 * <dwarf-gnu@augmentedreality.de>
 *
 * PoseEkf.h
 * Daniel Pustka <pustka@in.tum.de> May 2004
 * Distributed Wearable Augmented Reality Framework - www.augmentedreality.de
 * For documentation and instructions, see the DWARF documentation website at
 * http://wwwbruegge.in.tum.de/projects/dwarf/doc/
 *
 * For further questions about DWARF, please contact us at
 * <dwarf@augmentedreality.de>
 * For questions about this particular service, contact
 * Daniel Pustka <pustka@in.tum.de>
 *
 * $Id: PoseEkf.cpp,v 1.1 2004/06/07 13:34:29 pustka Exp $
 */

#include "PoseEkf.h"
#include "EkfPredict.h"
#include "EkfMeasure.h"
#include <debug.h>
#include <TimeUtil.h>

using std::pair;
using std::list;
using std::auto_ptr;
typedef TNT::Array2D<double> Matrix;
typedef TNT::Array1D<double> Vector;
using TNT::matmult;
using namespace DWARF;


////////////////////////////////////////////////////////////////////////////
// TNT fixes

// TNT is crap. There is no matrix transpose.
static Matrix transpose( const Matrix& M )
   {
   int n = M.dim1();
   int m = M.dim2();

   Matrix Mt( m, n );

   for ( int i = 0; i < n; i++ )
      for ( int j = 0; j < m; j++ )
         Mt[j][i] = M[i][j];

   return Mt;
   }


////////////////////////////////////////////////////////////////////////////
// PoseEKF::MotionModel

PoseEKF::MotionModel::MotionModel( int nPosDerivations, int nOriDerivations )
   {
   SetPosDerivations( nPosDerivations );
   SetOriDerivations( nOriDerivations );
   }


void PoseEKF::MotionModel::SetPosDerivations( int nPosDerivations )
   {
   if ( nPosDerivations * 3 + 3 > static_cast<int>( m_PosProcessNoise.size() ) )
      m_PosProcessNoise.resize( nPosDerivations * 3 + 3, 1.0 );
   m_nPosDerivations = nPosDerivations;
   }


void PoseEKF::MotionModel::SetOriDerivations( int nOriDerivations )
   {
   if ( nOriDerivations * 3 + 4 > static_cast<int>( m_OriProcessNoise.size() ) )
      m_OriProcessNoise.resize( nOriDerivations >= 0 ? 4 + 3 * nOriDerivations : 0, 1.0 );
   m_nOriDerivations = nOriDerivations;
   }


////////////////////////////////////////////////////////////////////////////
// PoseEKF

PoseEKF::PoseEKF( const PoseEKF::MotionModel& MoMo )
   : ExtendedKalmanFilter( 3 + 3 * MoMo.GetPosDerivations() +
      ( MoMo.GetOriDerivations() >= 0 ? 4 + 3 * MoMo.GetOriDerivations() : 0 ) )
   , m_MoMo( MoMo )
   , m_nPosStateSize( 3 + 3 * m_MoMo.GetPosDerivations() )
   , m_bInitialized( false )
   {
   // position
   if ( m_MoMo.GetPosDerivations() >= 0 )
      AddTimeUpdateFunction(
         pair<int, int>( 0, 3 ),
         pair<int, int>( 0, 3 + 3 * m_MoMo.GetPosDerivations() ),
         std::auto_ptr<ExtendedKalmanFilter::TimeUpdateFunction>(
            new PolynomialPredictor( 3, m_MoMo.GetPosDerivations() ) ) );

   // position derivations
   for ( int i = 1; i <= m_MoMo.GetPosDerivations(); i++ )
      AddTimeUpdateFunction(
         pair<int, int>( 3 * i, 3 ),
         pair<int, int>( 3 * i, 3 * ( m_MoMo.GetPosDerivations() - i + 1 ) ),
         std::auto_ptr<ExtendedKalmanFilter::TimeUpdateFunction>(
            new PolynomialPredictor( 3, m_MoMo.GetPosDerivations() - i ) ) );

   // orientation
   if ( m_MoMo.GetOriDerivations() >= 0 )
      AddTimeUpdateFunction(
         pair<int, int>( m_nPosStateSize, 4 ),
         pair<int, int>( m_nPosStateSize, 4 + m_MoMo.GetOriDerivations() * 3 ),
         std::auto_ptr<ExtendedKalmanFilter::TimeUpdateFunction>(
            new QuaternionPredictor( m_MoMo.GetOriDerivations() ) ) );

   // orientation derivations
   for ( int i = 1; i <= m_MoMo.GetOriDerivations(); i++ )
      AddTimeUpdateFunction(
         pair<int, int>( m_nPosStateSize + 1 + 3 * i, 3 ),
         pair<int, int>( m_nPosStateSize + 1 + 3 * i, 3 * ( m_MoMo.GetOriDerivations() - i + 1 ) ),
         std::auto_ptr<ExtendedKalmanFilter::TimeUpdateFunction>(
            new PolynomialPredictor( 3, m_MoMo.GetOriDerivations() - i ) ) );

   // set process noise
   std::auto_ptr<double> _pPN( new double[ GetStateSize() ] );
   double* pPN = _pPN.get();

   // position PN
   for ( int i = 0; i < m_nPosStateSize; i++ )
      pPN[ i ] = m_MoMo.GetPosProcessNoise( i % 3, i / 3 );

   // orientation PN
   if ( m_MoMo.GetOriDerivations() >= 0 )
      {
      for ( int i = 0; i < 4; i++ )
         pPN[ m_nPosStateSize + i ] = m_MoMo.GetOriProcessNoise( 0 );
      for ( int i = 1; i < m_MoMo.GetOriDerivations() + 1; i++ )
         for ( int j = 0; j < 3; j++ )
            pPN[ m_nPosStateSize + 1 + 3 * i + j ] = m_MoMo.GetOriProcessNoise( i );
      }

   SetProcessNoise( pPN );
   }


void PoseEKF::Init( const DWARF::PoseData& P )
   {
   Vector NewState( m_nStateSize );
   Vector NewSigma( m_nStateSize );
   NewState = 0.0;
   NewSigma = 10.0;

   // position
   if ( m_MoMo.GetPosDerivations() >= 0 )
      {
      // copy state
      for ( int i = 0; i < 3; i++ )
         NewState[ i ] = P.position[ i ];

      // initialize covariance from trace of covariance
      double fPosTrace = 0.0;
      for ( int i = 0; i < 3; i++ )
         fPosTrace += P.covariance[ i ][ i ];
      fPosTrace = sqrt( fPosTrace );
      for ( int i = 0; i < 3; i++ )
         NewSigma[ i ] = fPosTrace;
      }

   if ( m_MoMo.GetOriDerivations() >= 0 )
      {
      // copy state
      for ( int i = 0; i < 4; i++ )
         NewState[ m_nPosStateSize + i ] = P.orientation[ i ];

      // initialize covariance from trace of covariance
      double fOriTrace = 0.0;
      for ( int i = 0; i < 3; i++ )
         fOriTrace += P.covariance[ 3 + i ][ 3 + i ];
      fOriTrace = sqrt( fOriTrace );
      for ( int i = 0; i < 4; i++ )
         NewSigma[ m_nPosStateSize + i ] = fOriTrace;
      }

   // init
   Initialize( NewState, NewSigma );

   // copy time
   m_Time = P.timestamp;

   m_bInitialized = true;
   }


void PoseEKF::AddPose( const DWARF::PoseData& P, int nPosDerivation, int nOriDerivation )
   {
   if ( !m_bInitialized )
      {
      Init( P );
      return;
      }

   if ( P.timestamp < m_Time )
      {
      DEBUGSTREAM( 1, "Cannot yet integrate late measurements!" );
      return;
      }

   try
      {
      // Time update
      Time TDiff = P.timestamp - m_Time;
      TimeUpdate( TDiff.seconds + TDiff.microseconds / 1000000.0 );
      m_Time = P.timestamp;

      // position and orientation
      if ( m_MoMo.GetPosDerivations() >= nPosDerivation && P.hasPosition &&
         m_MoMo.GetOriDerivations() >= nOriDerivation && P.hasOrientation )
         {
         // create measurement vector
         double fMeasurement[ 7 ];
         for ( int i = 0; i < 3; i++ )
            fMeasurement[ i ] = P.position[ i ];
         for ( int i = 0; i < 4; i++ )
            fMeasurement[ 3 + i ] = P.orientation[ i ];

         if ( nOriDerivation == 0 )
            {
            DWARF::Util::quaternionNegateIfCloser( fMeasurement + 3, &m_State[ m_nPosStateSize ] );

            // create measurement covariance matrix
            Matrix R = Pose2KFCov( P, true );

            // measure
            Measure( 7, pair<int, int>( 0, GetStateSize() ),
               CombinedMeasurement(
                  3, pair<int, int>( 3 * nPosDerivation, 4 ),
                     auto_ptr<ExtendedKalmanFilter::MeasurementFunction>( new IdentityMeasurement( 3 ) ),
                  4, pair<int, int>( m_nPosStateSize, 4 ),
                     auto_ptr<ExtendedKalmanFilter::MeasurementFunction>( new QuaternionMeasurement ) ),
               fMeasurement, R );
            }
         else
            {
            // create measurement covariance matrix
            const Matrix R( 6, 6, const_cast<double*>( P.covariance[ 0 ] ) );

            // measure
            Measure( 6, pair<int, int>( 0, GetStateSize() ),
               CombinedMeasurement(
                  3, pair<int, int>( 3 * nPosDerivation, 4 ),
                     auto_ptr<ExtendedKalmanFilter::MeasurementFunction>( new IdentityMeasurement( 3 ) ),
                  3, pair<int, int>( m_nPosStateSize + 1 + 3 * nOriDerivation, 3 ),
                     auto_ptr<ExtendedKalmanFilter::MeasurementFunction>( new IdentityMeasurement( 3 ) ) ),
               fMeasurement, R );
            }
         }
      // position only
      else if ( m_MoMo.GetPosDerivations() >= nPosDerivation && P.hasPosition )
         {
         // create measurement covariance matrix
         // TNT is crap: fix subarray!
         Matrix R( 3, 3 );
         for ( int i = 0; i < 3; i++ )
            for( int j = 0; j < 3; j++ )
               R[ i ][ j ] = P.covariance[ i ][ j ];

         // measure
         Measure( 3, pair<int, int>( 3 * nPosDerivation, 3 ), IdentityMeasurement( 3 ), P.position, R );
         }
      // orientation only
      else if ( m_MoMo.GetOriDerivations() >= nOriDerivation && P.hasOrientation )
         {
         if ( nOriDerivation == 0 )
            {
            // create measurement vector
            double fMeasurement[ 4 ];
            for ( int i = 0; i < 4; i++ )
               fMeasurement[ i ] = P.orientation[ i ];
            DWARF::Util::quaternionNegateIfCloser( fMeasurement, &m_State[ m_nPosStateSize ] );

            // create measurement covariance matrix
            Matrix R = Pose2KFCov( P, false );

            // measure
            Measure( 4, pair<int, int>( m_nPosStateSize, 4 ), QuaternionMeasurement(), fMeasurement, R );
            }
         else
            {
            // create measurement covariance matrix
            // TNT is crap: fix subarray!
            Matrix R( 3, 3 );
            for ( int i = 0; i < 3; i++ )
               for( int j = 0; j < 3; j++ )
                  R[ i ][ j ] = P.covariance[ 3 + i ][ 3 + j ];

            // measure
            Measure( 3, pair<int, int>( m_nPosStateSize + 1 + 3 * nOriDerivation, 3 ),
               IdentityMeasurement( 3 ), P.orientation, R );
            }
         }

      // normalize quaternion
      if ( m_MoMo.GetOriDerivations() >= nOriDerivation && P.hasOrientation )
         DWARF::Util::normalizeQuaternion( &m_State[ m_nPosStateSize ] );

      }
   catch( ... )
      {
      DEBUGSTREAM( 1, "Got unstable - resetting EKF" );
      m_bInitialized = false;
      }
   }


DWARF::PoseData PoseEKF::PredictPose( const DWARF::Time& PTime )
   {
   // initialize a PoseData
   PoseData P = EasyPoseData();
   P.timestamp = PTime;

   // compute time delta
   Time TDiff;
   double fDeltaTime = 1.0;
   if ( PTime < m_Time )
      {
      TDiff = m_Time - PTime;
      fDeltaTime = -1.0;
      }
   else
      TDiff = PTime - m_Time;

   fDeltaTime *= TDiff.seconds + TDiff.microseconds / 1000000.0;

   // setup prediction list
   std::list< std::pair< int, int > > Pl;
   if ( m_MoMo.GetPosDerivations() >= 0 )
      Pl.push_back( std::pair< int, int >( 0, 3 ) );
   if ( m_MoMo.GetOriDerivations() >= 0 )
      Pl.push_back( std::pair< int, int >( m_nPosStateSize, 4 ) );

   // get position
   double fResult[ 7 ];
   Matrix R = Predict( fDeltaTime, Pl, fResult );

   if ( m_MoMo.GetPosDerivations() >= 0 && m_MoMo.GetOriDerivations() >= 0 )
      {
      P.hasPosition = P.hasOrientation = true;
      for ( int i = 0; i < 3; i++ )
         P.position[ i ] = fResult[ i ];
      for ( int i = 0; i < 4; i++ )
         P.orientation[ i ] = fResult[ 3 + i ];

      Matrix V = GetV( fResult + 3, true );
      R = matmult( matmult( transpose( V ), R ), V );

      for ( int i = 0; i < 6; i++ )
         for ( int j = 0; j < 6; j++ )
            P.covariance[ i ][ j ] = R[ i ][ j ];
      }
   else if ( m_MoMo.GetPosDerivations() >= 0 )
      {
      P.hasPosition = true;
      for ( int i = 0; i < 3; i++ )
         P.position[ i ] = fResult[ i ];
      for ( int i = 0; i < 3; i++ )
         for ( int j = 0; j < 3; j++ )
            P.covariance[ i ][ j ] = R[ i ][ j ];
      }
   else if ( m_MoMo.GetOriDerivations() >= 0 )
      {
      P.hasOrientation = true;
      for ( int i = 0; i < 4; i++ )
         P.orientation[ i ] = fResult[ i ];
      Matrix V = GetV( fResult, false );
      R = matmult( matmult( transpose( V ), R ), V );

      for ( int i = 0; i < 3; i++ )
         for ( int j = 0; j < 3; j++ )
            P.covariance[ 3 + i ][ 3 + j ] = R[ i ][ j ];
      }

   return P;
   }


TNT::Array2D<double> PoseEKF::GetV( const double* pOrientation, bool bWithPos )
   {
   Matrix R;
   int i;
   if ( bWithPos )
      {
      R = Matrix( 7, 7 );
      R = 0.0;
      R[0][0] = R[1][1] = R[2][2] = 1.0;
      i = 3;
      }
   else
      {
      R = Matrix( 4, 4 );
      i = 0;
      }

   double x = pOrientation[ 0 ];
   double y = pOrientation[ 1 ];
   double z = pOrientation[ 2 ];
   double w = pOrientation[ 3 ];
   R[i  ][i  ] =  w; R[i  ][i+1] = -z; R[i  ][i+2] =  y; R[i  ][i+3] =  x;
   R[i+1][i  ] =  z; R[i+1][i+1] =  w; R[i+1][i+2] = -x; R[i+1][i+3] =  y;
   R[i+2][i  ] = -y; R[i+2][i+1] =  x; R[i+2][i+2] =  w; R[i+2][i+3] =  z;
   R[i+3][i  ] = -x; R[i+3][i+1] = -y; R[i+3][i+2] = -z; R[i+3][i+3] =  w;

   return R;
   }


TNT::Array2D<double> PoseEKF::Pose2KFCov( const DWARF::PoseData& P, bool bWithPos )
   {
   Matrix R;
   if ( bWithPos )
      {
      R = Matrix( 7, 7 );

      // TNT is crap: fix subarray!
      for ( int i = 0; i < 6; i++ )
         for( int j = 0; j < 6; j++ )
            R[ i ][ j ] = P.covariance[ i ][ j ];
      for ( int i = 0; i < 6; i++ )
         R[ 6 ][ i ] = R[ i ][ 6 ] = 0.0;
      R[ 6 ][ 6 ] = 1.0;
      }
   else
      {
      R = Matrix( 4, 4 );

      // TNT is crap: fix subarray!
      for ( int i = 0; i < 3; i++ )
         for( int j = 0; j < 3; j++ )
            R[ i ][ j ] = P.covariance[ 3 + i ][ 3 + j ];
      for ( int i = 0; i < 3; i++ )
         R[ 3 ][ i ] = R[ i ][ 3 ] = 0.0;
      R[ 3 ][ 3 ] = 1.0;
      }

   Matrix V = GetV( P.orientation, bWithPos );
   R = matmult( matmult( V, R ), transpose( V ) );
   return R;
   }

