/**
 * @file dwarfutil.cpp
 * @author Martin Wagner
 *
 * @brief Implementation of some often used utility functions
 *
 * $Id: dwarfutil.cpp,v 1.28 2004/12/09 00:53:15 siggelko Exp $
 * $Revision: 1.28 $
 */

#include <config.h>
#include <debug.h>

// DWARF uses NODEBUG, glibc uses NDEBUG...
#ifdef NODEBUG
#define NDEBUG
#endif

#include <assert.h>

// DWARF uses NODEBUG, glibc uses NDEBUG...
#ifdef NODEBUG
#undef NDEBUG
#endif



#ifdef WIN32
# pragma warning(disable: 4786)
# include <windows.h>
# include <time.h>
#else
# include <unistd.h>
#endif


#include <math.h>


#ifdef HAVE_GETTIMEOFDAY
# include <time.h>
# ifdef HAVE_SYS_TIME_H
#  ifdef TIME_WITH_SYS_TIME
#   include <sys/time.h>
#  endif
# endif
#else
# include <sys/timeb.h>
#endif


#include "dwarfutil.h"


using namespace DWARF;
using namespace DWARF::Util;



//! Tolerance for quaternion operations
const double Q_EPSILON = (1e-10);
//! Tolerance for unit quaternion checking
const double Q_UNITCHECK = (1e-5);

//! macro to check whether quaternion is a unit quaternion
#define checkUnit(q) \
	assert( fabs(q[QX]*q[QX] + q[QY]*q[QY] + q[QZ]*q[QZ] + q[QW]*q[QW] - 1) <= Q_UNITCHECK)





Time DWARF::Util::currentTime() {

   Time now;

#ifdef _MSC_VER

   // Windows time
   struct _timeb wintime;
   _ftime( &wintime );
   now.seconds = wintime.time;
   now.microseconds = wintime.millitm * 1000;

#else

   // Unix time
   struct timeval tvnow;
   gettimeofday( &tvnow, 0 );
   now.seconds = tvnow.tv_sec;
   now.microseconds = tvnow.tv_usec;

#endif

   return now;
}

DWARF::Time DWARF::Util::getTimeToNow(Time &compareTime)
{
	Time now = currentTime();
	Time retVal;
	retVal.seconds = now.seconds - compareTime.seconds;
	if( now.microseconds < compareTime.microseconds ) {
		retVal.seconds--;
		retVal.microseconds = now.microseconds + ( 1000000 - compareTime.microseconds );
	}
	else
		retVal.microseconds = now.microseconds - compareTime.microseconds;

	return retVal;
}

/**
 * implementation based on precomputation, correct mult
 * matrix can be found in Horn, Closed-form solution of
 * absolute orientation using unit quaternions. (1987)
 */
double* DWARF::Util::quaternionToMatrix( const double *q, double *m)
{
	checkUnit(q);
	// precomputation of some values
	double xy = q[QX]*q[QY];
	double xz = q[QX]*q[QZ];
	double yz = q[QY]*q[QZ];
	double ww = q[QW]*q[QW];
	double wx = q[QW]*q[QX];
	double wy = q[QW]*q[QY];
	double wz = q[QW]*q[QZ];

	// row determining new x axis
	m[0] = 2*(q[QX]*q[QX] + ww) - 1;
	m[1] = 2 * (xy - wz);
	m[2] = 2 * (wy + xz);
	// row determining new y axis
	m[4] = 2 * (xy + wz);
	m[5] = 2*(q[QY]*q[QY] + ww) - 1;
	m[6] = 2 * (yz - wx);
	// row determining new z axis
	m[8] = 2 * (xz - wy);
	m[9] = 2 * (wx + yz);
	m[10] = 2*(q[QZ]*q[QZ] + ww) - 1;

	// fill matrix to make sure it's a valid homogeneous 4x4 matrix
	m[3] = m[7] = m[11] = 0.0;
	m[12] = m[13] = m[14] = 0.0;
	m[15] = 1.0;

	return m;
}

/**
 * implementation based on proposal by Horn, idea:
 * first determine the largest entry of unit quaternion,
 * then use this to get other entries
 */
double* DWARF::Util::matrixToQuaternion( const double *m, double *q )
{
	// get entry of q with largest absolute value
	// note: we compute here 4 * q[..]^2 - 1
	double tmp[4];
	tmp[QW] = m[0] + m[5] + m[10];
	tmp[QX] = m[0] - m[5] - m[10];
	tmp[QY] = -m[0] + m[5] - m[10];
	tmp[QZ] = -m[0] - m[5] + m[10];
	int max = QW;
	if( tmp[QX] > tmp[max] ) max = QX;
	if( tmp[QY] > tmp[max] ) max = QY;
	if( tmp[QZ] > tmp[max] ) max = QZ;

	// depending on largest entry compute the other values
	// note: these formulae can be derived very simply from the
	//       matrix representation computed in quaternionToMatrix
	switch(max) {
	case QW:
		q[QW] = sqrt(tmp[QW]+1) * 0.5;
		q[QX] = (m[9] - m[6]) / ( 4 * q[QW] );
		q[QY] = (m[2] - m[8]) / ( 4 * q[QW] );
		q[QZ] = (m[4] - m[1]) / ( 4 * q[QW] );
		break;
	case QX:
		q[QX] = sqrt(tmp[QX]+1) * 0.5;
		q[QW] = (m[9] - m[6]) / ( 4 * q[QX] );
		q[QY] = (m[4] + m[1]) / ( 4 * q[QX] );
		q[QZ] = (m[2] + m[8]) / ( 4 * q[QX] );
		break;
	case QY:
		q[QY] = sqrt(tmp[QY]+1) * 0.5;
		q[QW] = (m[2] - m[8]) / ( 4 * q[QY] );
		q[QX] = (m[4] + m[1]) / ( 4 * q[QY] );
		q[QZ] = (m[9] + m[6]) / ( 4 * q[QY] );
		break;
	case QZ:
		q[QZ] = sqrt(tmp[QZ]+1) * 0.5;
		q[QW] = (m[4] - m[1]) / ( 4 * q[QZ] );
		q[QX] = (m[2] + m[8]) / ( 4 * q[QZ] );
		q[QY] = (m[9] + m[6]) / ( 4 * q[QZ] );
		break;
	default:
		DEBUGSTREAM(0, "Error in switch clause. This should never happen.");
		assert(0);
		break;
	}

	normalizeQuaternion(q);

	return q;
}


double* DWARF::Util::axisAngleToQuaternion( double* axisAngle, double* quat )
{
	double length = sqrt( axisAngle[0]*axisAngle[0] + axisAngle[1]*axisAngle[1]
												+ axisAngle[2]*axisAngle[2] );
	if( length < Q_EPSILON ) {
		quat[QW] = 0; quat[QX] = 0; quat[QY] = 1.0; quat[QZ] = 0;
		return quat;
	}
	double ahalf = axisAngle[3]/2.0;
	double s = sin( ahalf );
	quat[QW] = cos( ahalf );
	quat[QX] = s * axisAngle[0] / length;
	quat[QY] = s * axisAngle[1] / length;
	quat[QZ] = s * axisAngle[2] / length;

	checkUnit(quat);

	return quat;
}


double* DWARF::Util::quaternionToAxisAngle( double* q, double* axisAngle )
{
	checkUnit(q);

	axisAngle[0] = q[QX];
	axisAngle[1] = q[QY];
	axisAngle[2] = q[QZ];
	axisAngle[3] = 2 * acos( q[QW] );

	return axisAngle;
}



double* DWARF::Util::invertQuaternion( double *q )
{
	double norm = 1.0 / (q[QX]*q[QX] + q[QY]*q[QY] + q[QZ]*q[QZ] + q[QW]*q[QW]);
	q[QX] *= -norm;
	q[QY] *= -norm;
	q[QZ] *= -norm;
	return q;
}


double* DWARF::Util::conjugateQuaternion( double *q )
{
	q[QX] = -q[QX];
	q[QY] = -q[QY];
	q[QZ] = -q[QZ];
	return q;
}


double* DWARF::Util::normalizeQuaternion( double *q ) 
{
	return normalizeVector(q, 4);
}

double* DWARF::Util::normalizeVector( double *vec, int length )
{
	double norm = 0.0;
	for(int i=0; i<length; i++) norm += vec[i]*vec[i];
	norm = sqrt( 1.0/ norm );
	for(int i=0; i<length; i++) vec[i] *= norm;

	return vec;
}

double* DWARF::Util::multiplyQuaternion( double *q, const double *q1, const double *q2 )
{
	q[QX] = q1[QW]*q2[QX] + q1[QX]*q2[QW] + q1[QY]*q2[QZ] - q1[QZ]*q2[QY];
	q[QY] = q1[QW]*q2[QY] - q1[QX]*q2[QZ] + q1[QY]*q2[QW] + q1[QZ]*q2[QX];
	q[QZ] = q1[QW]*q2[QZ] + q1[QX]*q2[QY] - q1[QY]*q2[QX] + q1[QZ]*q2[QW];
	q[QW] = q1[QW]*q2[QW] - q1[QX]*q2[QX] - q1[QY]*q2[QY] - q1[QZ]*q2[QZ];
	return q;
}


/**
 * implementation based on precomputation, correct mult
 * matrix can be found in Horn, Closed-form solution of
 * absolute orientation using unit quaternions. (1987)
 */
double* DWARF::Util::rotateQuaternion( double *r, const double *q, const double *p )
{
	checkUnit(q);

	// precomputation of some values
	double xy = q[QX]*q[QY];
	double xz = q[QX]*q[QZ];
	double yz = q[QY]*q[QZ];
	double ww = q[QW]*q[QW];
	double wx = q[QW]*q[QX];
	double wy = q[QW]*q[QY];
	double wz = q[QW]*q[QZ];

	r[0] = p[0] * ( 2*(q[QX]*q[QX] + ww) - 1 ) + p[1] * 2 * (xy - wz) + p[2] * 2 * (wy + xz);
	r[1] = p[0] * 2 * (xy + wz) + p[1] * ( 2*(q[QY]*q[QY] + ww) - 1 ) + p[2] * 2 * (yz - wx);
	r[2] = p[0] * 2 * (xz - wy) + p[1] * 2 * (wx + yz) + p[2] * ( 2*(q[QZ]*q[QZ] + ww) - 1 );

	return r;
}

void DWARF::Util::transformCoordinateFrame( const double *transRot, const double *transPos,
	double *rot, double *pos)
{
	// we compute (in homogeneous representation, P'=(transRot, transPos) and P=(rot, pos) )
	// P'^-1 * P * P'
	// thus, rot_new = transRot^-1 * rot * transRot, pos_new = transRot^-1 * pos
	double tmpRot[4], invTransRot[4];
	invTransRot[0] = transRot[0];
	invTransRot[1] = transRot[1];
	invTransRot[2] = transRot[2];
	invTransRot[3] = transRot[3];
	invertQuaternion( invTransRot );
	// rotation first
	multiplyQuaternion( tmpRot, invTransRot, rot );
	multiplyQuaternion( rot, tmpRot, transRot );
	// position last
	double tmpPos[3];
	rotateQuaternion( tmpPos, invTransRot, pos );
	pos[0] = tmpPos[0];
	pos[1] = tmpPos[1];
	pos[2] = tmpPos[2];

	return;
}

double* DWARF::Util::rotateMatrix( double* rotPoint, const double *matrix, const double *point )
{
	int row_index = 0;
	for(int row=0; row<3; row++) {
		rotPoint[row] = matrix[row_index] * point[0] + matrix[row_index+1] * point[1] +
                        matrix[row_index+2] * point[2] + matrix[row_index+3];
		row_index += 4;
	}
	return rotPoint;
}

double *DWARF::Util::invertMatrix( double *ap, int col, int row )
{
        double *wap, *wcp, *wbp;/* work pointer                 */
        int i,j,n,ip,nwork;
        int nos[50];
        double epsl;
        double p,pbuf,work;
        //double  fabs();

        epsl = (double)1.0e-10;         /* Threshold value      */

        switch (col) {
                case (0): return(NULL);                 /* check size */
                case (1): *ap = (double)1.0 / (*ap);
                          return(ap);                   /* 1 dimension */
        }

        for(n = 0; n < col ; n++)
                nos[n] = n;

        for(n = 0; n < col ; n++) {
                wcp = ap + n * row;

                for(i = n, wap = wcp, p = 0.0; i < col ; i++, wap += row)
                        if( p < ( pbuf = (double)fabs(*wap)) ) {
                                p = pbuf;
                                ip = i;
                        }
                if (p <= epsl)
                        return(NULL);

                nwork = nos[ip];
                nos[ip] = nos[n];
                nos[n] = nwork;

                for(j = 0, wap = ap + ip * row, wbp = wcp; j < col ; j++) {
                        work = *wap;
                        *wap++ = *wbp;
                        *wbp++ = work;
                }

                for(j = 1, wap = wcp, work = *wcp; j < col ; j++, wap++)
                        *wap = *(wap + 1) / work;
                *wap = (double)1.0 / work;

                for(i = 0; i < col ; i++) {
                        if(i != n) {
                                wap = ap + i * row;
                                for(j = 1, wbp = wcp, work = *wap;
                                                j < col ; j++, wap++, wbp++)
                                        *wap = *(wap + 1) - work * (*wbp);
                                *wap = -work * (*wbp);
                        }
                }
        }

        for(n = 0; n < col ; n++) {
                for(j = n; j < col ; j++)
                        if( nos[j] == n) break;
                nos[j] = nos[n];
                for(i = 0, wap = ap + j, wbp = ap + n; i < col ;
                                        i++, wap += row, wbp += row) {
                        work = *wap;
                        *wap = *wbp;
                        *wbp = work;
                }
        }
        return(ap);
}

double* DWARF::Util::crossProduct3D( double *p, double *a, double *b )
{
	p[0] = a[1]*b[2] - a[2]*b[1];
	p[1] = a[2]*b[0] - a[0]*b[2];
	p[2] = a[0]*b[1] - a[1]*b[0];

	return p;
}

double DWARF::Util::scalarProduct3D( double *a, double *b )
{
	return a[0]*b[0]  + a[1]*b[1] + a[2]*b[2];
}

double DWARF::Util::angleBetweenVectors3D( double *a, double *b )
{
	normalizeVector(a,3);
	normalizeVector(b,3);
	double scalar = scalarProduct3D(a,b);
	
	return acos(scalar);
}

void DWARF::Util::poseRightInverse( double* newQuat, double* newTrans,
	const double* oldQuat, const double* oldTrans )
{
	checkUnit(oldQuat);

	newQuat[0] = oldQuat[0];
	newQuat[1] = oldQuat[1];
	newQuat[2] = oldQuat[2];
	newQuat[3] = oldQuat[3];
	conjugateQuaternion(newQuat);

	rotateQuaternion( newTrans, newQuat, oldTrans );
	newTrans[0] = -newTrans[0];
	newTrans[1] = -newTrans[1];
	newTrans[2] = -newTrans[2];

	return;
}

void DWARF::Util::poseLeftInverse( double* newQuat, double* newTrans,
	const double* oldQuat, const double* oldTrans )
{
	checkUnit(oldQuat);

	newQuat[0] = oldQuat[0];
	newQuat[1] = oldQuat[1];
	newQuat[2] = oldQuat[2];
	newQuat[3] = oldQuat[3];
	conjugateQuaternion(newQuat);

	rotateQuaternion( newTrans, oldQuat, oldTrans );
	newTrans[0] = -newTrans[0];
	newTrans[1] = -newTrans[1];
	newTrans[2] = -newTrans[2];

	return;
}

/*
 * This code was partially taken from OpenTracker, TU Vienna
 * http://studierstube.org/opentracker/
 */
double* DWARF::Util::colMatrixToEuler( double* colMatrix, double* angles )
{
  double sinPitch, cosPitch, sinRoll, cosRoll, sinYaw, cosYaw;

	sinPitch = -colMatrix[8]; //[2][0];
	cosPitch = sqrt(1 - sinPitch*sinPitch);

	if ( fabs(cosPitch) > Q_EPSILON )	{
		sinRoll = colMatrix[9] / cosPitch;  // [2][1]
		cosRoll = colMatrix[10] / cosPitch; // [2][2]
		sinYaw = colMatrix[4] / cosPitch; // [1][0]
		cosYaw = colMatrix[0] / cosPitch; // [0][0]
	}
	else {
		sinRoll = -colMatrix[6]; // [1][2]
		cosRoll = colMatrix[5]; // [1][1]
		sinYaw = 0;
		cosYaw = 1;
	}
	/* roll */
	angles[0] = atan2(sinRoll, cosRoll);
	/* pitch */
	angles[1] = atan2(sinPitch, cosPitch);
	/* yaw */
	angles[2] = atan2(sinYaw, cosYaw);

	return angles;
}


/*
 * This code was partially taken from OpenTracker, TU Vienna
 * http://studierstube.org/opentracker/
 */
double* DWARF::Util::eulerToQuaternion( double* angles, double* quat )
{
	double cr, cp, cy, sr, sp, sy, cpcy, spsy;

	// calculate trig identities
	cr = cos(angles[0]/2); // roll
	cp = cos(angles[1]/2); // pitch
	cy = cos(angles[2]/2); // yaw

	sr = sin(angles[0]/2); // roll
	sp = sin(angles[1]/2); // pitch
	sy = sin(angles[2]/2); // yaw

	cpcy = cp * cy;
	spsy = sp * sy;

	quat[QW] = (float)(sr * cpcy - cr * spsy);
	quat[QX] = (float)(cr * sp * cy + sr * cp * sy);
	quat[QY] = (float)(cr * cp * sy - sr * sp * cy);
	quat[QZ] = (float)(cr * cpcy + sr * spsy);
	return quat;
}


double* DWARF::Util::quaternionSlerp( double* quat, const double* quat1, const double* quat2, double h )
{
	// copy quat1 to q1 temp and do negateIfCloser
	double q1[ 4 ];
	for ( int i = 0; i < 4; i++ )
		q1[ i ] = quat1[ i ];
	quaternionNegateIfCloser( q1, quat2 );

	// compute dot-product
	double fDot = q1[ 0 ] * quat2[ 0 ] + q1[ 1 ] * quat2[ 1 ] + q1[ 2 ] * quat2[ 2 ] + q1[ 3 ] * quat2[ 3 ];

	double fW1, fW2;
	if ( fDot > 0.9999 )
	{
		// small angle => linear interpolation
		fW1 = 1.0 - h;
		fW2 = h;
	} else {
		// SLERP
		double fOmega = acos( fDot );
		double fSinOmega = sin( fOmega );
		fW1 = sin( ( 1.0 - h ) * fOmega ) / fSinOmega;
		fW2 = sin( h * fOmega ) / fSinOmega;
	}

	for ( int i = 0; i < 4; i++ )
		quat[ i ] = fW1 * q1[ i ] + fW2 * quat2[ i ];

	return normalizeQuaternion( quat ); // probably unnecessary
}


double* DWARF::Util::quaternionNegateIfCloser( double* quat, const double* otherquat )
{
	double fDiff = 0.0;
	for(int i = 0; i < 4; i++ )
		fDiff += ( quat[i] - otherquat[i] ) * ( quat[i] - otherquat[i] );

	if ( fDiff > 2.0 )
		negateQuaternion( quat );

	return quat;
}

double* DWARF::Util::negateQuaternion( double *quat )
{
	quat[QX] = -quat[QX];
	quat[QY] = -quat[QY];
	quat[QZ] = -quat[QZ];
	quat[QW] = -quat[QW];

	return quat;
}
