/**
 * @file dwarfutil.h
 *
 * Utility classes and functions to simplify working with the 
 * DWARF framework under C++. Mainly conversion routines between different
 * representations of rotations (axis/angle, quaternions, rotation matrices,
 * euler angles)
 *
 * @author Martin Wagner
 */

/*
 * Distributed Wearable Augmented Reality Framework - www.augmentedreality.de
 * $Id: dwarfutil.h,v 1.25 2004/12/09 00:53:15 siggelko Exp $
 * $Revision: 1.25 $
 */

#ifndef __DWARFUTIL_H_
#define __DWARFUTIL_H_

#include <DWARF/DwarfCommon.h>


namespace DWARF {
  namespace Util {

		/**
		 * @brief Get the current system time as a DWARF::Time struct
		 *
		 * @return: the current system time
		 */
		Time currentTime();

		/**
		 * @brief Get the difference to the current time
		 *
		 * @param compareTime the time to be be compared
		 */
		Time getTimeToNow(Time &compareTime);

		/*
		 * Some constants and methods for accessing quaternion components
		 */
		//! @brief Use only this constant for accessing a quaternion's X component
		const unsigned char QX = 0;
		//! @brief Use only this constant for accessing a quaternion's Y component
		const unsigned char QY = 1;
		//! @brief Use only this constant for accessing a quaternion's Z component
		const unsigned char QZ = 2;
		//! @brief Use only this constant for accessing a quaternion's scalar component
		const unsigned char QW = 3;
		//! @brief Convenience method for accessing a quaternion's X component
		inline double q_x( double *q) { return q[QX]; }
		//! @brief Convenience method for accessing a quaternion's y component
		inline double q_y( double *q) { return q[QY]; }
		//! @brief Convenience method for accessing a quaternion's z component
		inline double q_z( double *q) { return q[QZ]; }
		//! @brief Convenience method for accessing a quaternion's scalar component
		inline double q_w( double *q) { return q[QW]; }

		/**
		 * @brief Construct rotation matrix from unit quaternion.
		 *
		 * Assumes matrix is used to multiply column vector on the left:
		 * vnew = mat vold. Works correctly for right-handed coordinate system
		 * and right-handed rotations. 
		 *
		 * @param quat 4-dim unit quaternion as input
		 * @param matrix 4x4-dim double homogeneous rotation matrix
		 *                as return value, organized as follows (row major)<br>
		 *                <b>NOTE: OpenGL has another order, column major!</b><br>
		 * <CODE>           ( 0  1  2  3 )<br>
		 *                  ( 4  5  6  7 )<br>
		 *                  ( 8  9 10 11 )<br>
		 *                  (12 13 14 15 )
		 * </CODE>
		 * @return pointer to matrix
		 */
		double *quaternionToMatrix( const double *quat, double *matrix);

		/**
		 * @brief Construct a unit quaternion from a rotation matrix. 
		 * 
		 * Assumes matrix is 
		 * used to multiply column vector on the left: vnew = mat vold. Works 
		 * correctly for right-handed coordinate system and right-handed rotations. 
		 * Translation components are ignored. Matrix is assumed to be orthonormal.
		 *
		 * @param matrix 4x4-dim double homogeneous matrix
		 *                as input, organized as follows (row major)<br>
		 *                <b>NOTE: OpenGL has another order, column major!</b><br>
		 * <CODE>           ( 0  1  2  3 )<br>
		 *                  ( 4  5  6  7 )<br>
		 *                  ( 8  9 10 11 )<br>
		 *                  (12 13 14 15 )
		 * </CODE>
		 * @param quat 4-dim normalized quaternion as return value
		 * @return pointer to quat
		 */ 
		double *matrixToQuaternion( const double *matrix, double *quat );


		/**
		 * @brief Converts axis/angle representation (as used in VRML) to a unit quaternion
		 * 
		 * @param axisAngle 4-dim double axis/angle rotation (x,y,z,alpha) as input
		 * @param quat 4-dim unit quaternion as return value
		 * @return a pointer to quat
		 */
		double* axisAngleToQuaternion( double* axisAngle, double* quat );
		inline double* vrmlToQuaternion(double* vrml, double* q) {
			return axisAngleToQuaternion(vrml, q);
		}

		/**
		 * @brief Converts a unit quaternion to axis/angle representation (as used in VRML)
		 *
		 * @param quat 4-dim unit quaternion as input
		 * @param axisAngle 4-dim double axis/angle rotation (x,y,z,alpha)
		 *              as return value
		 * @return a pointer to axisAngle
		 */
		double* quaternionToAxisAngle( double* quat, double* axisAngle );
		inline double* quaternionToVrml(double* q, double* vrml) {
			return quaternionToAxisAngle(q, vrml);
		}

		/**
		 * @brief Inverts a quaternion.
		 *
		 * The resulting quaternion is the inverse of the given quaternion. This
		 * corresponds to translating an orthonormal rotation matrix.
		 * Note: To invert a unit quaternion, conjugating it suffices.
		 *
		 * @param quat 4-dim quaternion as input and return value
		 * @return a pointer to quat
		 */
		double* invertQuaternion( double *quat );

		/**
		 * @brief Conjugates a quaternion (i.e. negates its vector part)
		 *
		 * Note: Conjugating a unit quaternion results in its inverse.
		 *
		 * @param quat 4-dim quaternion as input and return value
		 * @return a pointer to quat
		 */
		double* conjugateQuaternion( double *quat );

		/**
		 * @brief Normalizes a quaternion (i.e. makes a unit quaternion)
		 *
		 * @param quat 4-dim quaternion as input and return value
		 * @return a pointer to quat
		 */
		double* normalizeQuaternion( double *quat );

		/**
		 * @brief Normalizes a vector
		 *
		 * @param vec n-dim vector as input and return value
		 * @param length dimension of vector
		 * @return a pointer to vec
		 */
		double* normalizeVector( double *vec, int length );

		/**
		 * multiplyQuaternion
		 * Calculates <CODE> quat := quat1 * quat2 </CODE>
		 *
		 * @param quat 4-dim Quaternion as return value
		 * @param quat1 4-dim Quaternion as input
		 * @param quat2 4-dim Quaternion as input
		 * @return a pointer to quat
		 */
		double* multiplyQuaternion( double *quat, const double *quat1, const double *quat2 );

		/**
		 * @brief Rotates a 3-dim point by a unit quaternion
		 *
		 * Calculates <CODE> rotPoint = quat * (point,0) * quat^-1 </CODE>
		 *
		 * @param quat 4-dim unit quaternion representing the rotation
		 * @param point 3-dim point to be rotated as input
		 * @param rotPoint memory for return value (at least 3*double)
		 * @return a pointer to rotPoint
		 */
		double* rotateQuaternion( double *rotPoint, const double *quat, const double *point );

		/**
		 * @brief Transforms the coordinate reference frame of a 6D pose.
		 *
		 * Use this method to transform the reference frame of a 6DOF pose.
		 * If given the transformation pose P' and an input pose P, it computes
		 * P'^-1 * P * P'
		 *
		 * @param transRot rotational part of transformation P', 4-double unit quaternion
		 * @param transPos positional part of transformation P', 3-double position
		 * @param rot rotational part of input pose P, is modified, 4-double unit quaternion
		 * @param pos positional part of input pose P, is modified, 3-double position
		 */
		void transformCoordinateFrame( const double *transRot, const double *transPos,
			double *rot, double *pos);

		/**
		 * @brief Rotates a 3-dim point by a orthonormal rotation matrix
		 *
		 * Calculates <CODE> rotPoint = matrix * (point,1) </CODE>
		 *
		 * @param quat 3x4-dim row major orthonormal matrix representing the rotation, will not be changed
		 * @param point 3-dim point to be rotated as input
		 * @param rotPoint memory for return value (at least 3*double)
		 * @return a pointer to rotPoint
		 */
		double* rotateMatrix( double *rotPoint, const double *matrix, const double *point );

		/**
		 * @brief (self) inverts a matrix
		 *
		 * Calculates <CODE> matrix = matrix^-1 </CODE>
		 *
		 * @param matrix the matix that should be inverted
		 * @param col count of the colums in the matix
		 * @param row count of the rows in the matix
		 * @return a pointer to the inverted matix
		 */
		double* invertMatrix( double *matrix, int col, int row );


		/**
		 * @brief Compute cross product p = a x b of two 3D-vectors
		 *
		 * @param p 3-dim pointer to return value memory
		 * @param p 3-dim vector a
		 * @param p 3-dim vector b
		 *
		 * @return a pointer to p = a x b
		 */
		double* crossProduct3D( double *p, double *a, double *b );

		/**
		 * @brief Compute scalar product p = a . b of two 3D-vectors
		 *
		 * @param a 3-dim vector a
		 * @param b 3-dim vector b
		 *
		 * @return the scalar value p = a . b
		 */
		double scalarProduct3D(  double *a, double *b );
		
		/**
		 * @brief calculates the angle between two 3D-vectors
		 *
		 * @param a 3-dim vector a
		 * @param b 3-dim vector b
		 *
		 * @return the angle
		 */
		double angleBetweenVectors3D(  double *a, double *b );

		/**
		 * @brief Inverts a pose transformation.
		 *
		 * Identical to inverting a homogeneous 4x4 matrix, computes
		 * rot_new = rot_old^-1; t_new = - rot_old^-1 * t_old
		 *
		 * This function computes the right inverse, i.e. H^-1 such that
		 * H* H^-1 = I
		 *
		 * @param newQuat 4-dim unit quaternion, new rotation (return value)
		 * @param newTrans 3-dim pont, new translation (return value)
		 * @param oldQuat 4-dim unit quaternion, old rotation
		 * @param oldTrans 3-dim point, old translation
		 */
		void poseRightInverse( double* newQuat, double* newTrans,
			const double* oldQuat, const double* oldTrans );
		inline void invertQuaternionPose( double* newQuat, double* newTrans,
			const double* oldQuat, const double* oldTrans ) {
			return poseRightInverse(newQuat, newTrans, oldQuat, oldTrans);
		}

		/**
		 * @brief Inverts a pose transformation.
		 *
		 * Identical to inverting a homogeneous 4x4 matrix, computes
		 * rot_new = rot_old^-1; t_new = - rot_old * t_old
		 *
		 * This function computes the left inverse, i.e. H^-1 such that
		 * H^-1 * H = I
		 *
		 * @param newQuat 4-dim unit quaternion, new rotation (return value)
		 * @param newTrans 3-dim pont, new translation (return value)
		 * @param oldQuat 4-dim unit quaternion, old rotation
		 * @param oldTrans 3-dim point, old translation
		 */
		void poseLeftInverse( double* newQuat, double* newTrans,
			const double* oldQuat, const double* oldTrans );

		/**
		 * @brief Contructs Euler rotation representation out of matrix
		 *
		 * This code was partially taken from OpenTracker, TU Vienna
		 * http://studierstube.org/opentracker/
		 * originally, it seems to be from Gary Bishop
		 *
		 * FIXME: This code is supposed to take a column matrix. Ask
		 *        Gerhard Reitmayr whether that is the same as the
		 *        representation below or not. Adjust comment if
		 *        necessary. DO NOT change the code in here, it is
		 *        for the ARTkPoseReconstruction service.
		 *
		 * @param colMatrix: 4x4-dim double homogeneous matrix (row x column)
		 *                   as input, organized as follows<br>
		 * <CODE>           ( 0  1  2  3 )<br>
		 *                  ( 4  5  6  7 )<br>
		 *                  ( 8  9 10 11 )<br>
		 *                  (12 13 14 15 )
		 * </CODE>
		 * @param angles 3-dim angles (roll, pitch, yaw) as return value
		 * @return pointer to angles
		 */
		double* colMatrixToEuler( double* colMatrix, double* angles );

		/**
		 * @brief Constructs quaternion from Euler representation
		 *
		 * This code was partially taken from OpenTracker, TU Vienna
		 * http://studierstube.org/opentracker/
		 *
		 * @param angles 3-dim angles (roll, pitch, yaw)
		 * @param quat 4-dim normalized quaternion as return value
		 * @return pointer to quat
		 */
		double* eulerToQuaternion( double* angles, double* quat );

		/**
		 * @brief Interpolates quaternions using the SLERP algorithm
		 *
		 * @param quat destination quaternion
		 * @param quat1 first source quaternion
		 * @param quat2 second source quaternion
		 * @param h interpolation factor. 0.0=quat1, 1.0=quat2. May exceed [0,1] for extrapolation.
		 * @return pointer to quat
		 */
		double* quaternionSlerp( double* quat, const double* quat1, const double* quat2, double h );

		/**
		 * @brief Negates a quaternion if the result is closer to a given reference.
		 *
		 * Every rotation can be represented by two different quaternions, which
		 * can cause great trouble when analyzing rotation sequences. One can
		 * switch between the two representation by negating all four quaternion
		 * entries. This routine negates the quaternion if the result is closer to
		 * a given reference quaternion, e.g. a preceding quaternion in a sequence.
		 *
		 * @param quat Quaternion to be negated
		 * @param otherquat Reference quaternion
		 * @return pointer to quat
		 */
		double* quaternionNegateIfCloser( double* quat, const double* otherquat );

		/**
		 * @brief Negates a quaternion.
		 *
		 * Sets x = -x; y=-y; z=-z; w=-w
		 * The resulting quaternion still represents the same rotation.
		 * @param quat Quaternion to be negated
		 @return pointer to quat
		 */
		double* negateQuaternion( double *quat );
  };
};



#endif
