/***************************************************************************
 *                         EasyPoseData.cpp  -  description
 *                            -------------------
 *   begin                : 2004/03/31
 *   copyright            : (C) 2004 by Daniel Pustka
 *   email                : pustka@in.tum.de
 *
 *   $Id: EasyPoseData.cpp,v 1.9 2004/11/28 16:48:41 troels Exp $
 *   $Revision: 1.9 $
 *
  ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "EasyPoseData.h"

#include <tnt.h>
#include <cmath>
#include <TimeUtil.h>
#include <iomanip>

// Win32 Hack since std::isnan(x) is not a part of MSVC7 
//   but still lies in float.h as _isnan(x)
// If you know a better solution please change this!
// comments by Troels Frimor, troels@in.tum.de
#ifdef _MSC_VER 
  #include <float.h>
#endif

using namespace DWARF;

// forward declaration
static TNT::Array2D<double> generateInverseTransformMatrix( const PoseData& P );
static void generateProductTransformMatrices( const PoseData& P1, const PoseData& P2,
   TNT::Array2D<double>& J1, TNT::Array2D<double>& J2 );


// TNT is crap. There is no matrix transpose.
static TNT::Array2D<double> transpose( const TNT::Array2D<double>& M ) {
	int n = M.dim1();
	int m = M.dim2();

	TNT::Array2D<double> Mt( m, n );

	for ( int i = 0; i < n; i++ )
		for ( int j = 0; j < m; j++ )
			Mt[j][i] = M[i][j];

	return Mt;
}


// ensure that a symmetric matrix is symmetric
static void normalizeSymmetric( TNT::Array2D<double>& M ) {
	int n = M.dim1();

	for ( int i = 0; i < n - 1; i++ )
		for ( int j = i + 1; j < n; j++ )
			M[i][j] = M[j][i] = ( M[i][j] + M[j][i] ) / 2.0;
}


EasyPoseData::EasyPoseData( const char* pSource, const char* pTarget, bool bInitializeError )
  {
  hasPosition = false;
  hasOrientation = false;
  timestamp.seconds = 0;
  timestamp.microseconds = 0;
  confidence = 1.0;
  timeError = 1.0;

  if ( pSource )
    setSource( pSource );

  if ( pTarget )
    setTarget( pTarget );

  // initialize covariance with big errors
  if ( bInitializeError )
    for ( int i = 0; i < 6; i++ ) {
      for ( int j = 0; j < 6; j++ )
        covariance[ i ][ j ] = 0;
      covariance[ i ][ i ] = 1.0;
    }
}


void EasyPoseData::setPositionAccuracy( double sigma ) {
  for ( int i = 0; i < 3; i++ )
    for ( int j = 0; j < 3; j++ ) {
      covariance[ i ][ j ] = 0;
      // assume independence of position and orientation
      covariance[ i + 3 ][ j ] = 0;
      covariance[ i ][ j + 3 ] = 0;
    }

  covariance[0][0] = covariance[1][1] = covariance[2][2] = sigma * sigma;
}


void EasyPoseData::setOrientationAccuracy( double sigma ) {
  for ( int i = 0; i < 3; i++ )
    for ( int j = 0; j < 3; j++ ) {
      covariance[ i + 3 ][ j + 3 ] = 0;
      // assume independence of position and orientation
      covariance[ i + 3 ][ j ] = 0;
      covariance[ i ][ j + 3 ] = 0;
    }

  covariance[3][3] = covariance[4][4] = covariance[5][5] = sigma * sigma;
}


EasyPoseData EasyPoseData::inverse( const PoseData& P ) {
  // initialize return EasyPoseData
  EasyPoseData r( P.target.c_str(), P.source.c_str(), false );

  r.hasPosition = P.hasPosition;
  r.hasOrientation = P.hasOrientation;
  r.confidence = P.confidence;
  r.timestamp = P.timestamp;
  r.timeError = P.timeError;

  if ( P.hasPosition && P.hasOrientation ) {
    // we have both position and orientation
    DWARF::Util::invertQuaternionPose( r.orientation, r.position, P.orientation, P.position );

    // transform covariance
    TNT::Array2D<double> J = generateInverseTransformMatrix( P );
    TNT::Array2D<double> Jt = transpose( J );
    const TNT::Array2D<double> C( 6, 6, const_cast<double*>( P.covariance[0] ) );
    TNT::Array2D<double> Cnew( 6, 6, r.covariance[0] );
    Cnew.inject( TNT::matmult( TNT::matmult( J, C ), Jt ) );

  } else if ( P.hasPosition ) {
    // we only have position
    for ( int i = 0; i < 3; i++ )
      r.position[i] = -P.position[i];

    // just copy covariance
    for ( int i = 0; i < 3; i++ )
      for ( int j = 0; j < 3; j++ )
        r.covariance[i][j] = P.covariance[i][j];

  } else if ( P.hasOrientation ) {
    // we only have orientation
    for ( int i = 0; i < 3; i++ )
      r.orientation[i] = -P.orientation[i];
    r.orientation[3] = P.orientation[3];

    // just copy covariance
    for ( int i = 3; i < 6; i++ )
      for ( int j = 3; j < 6; j++ )
        r.covariance[i][j] = P.covariance[i][j];
  }

  return r;
}


EasyPoseData EasyPoseData::product( const PoseData& P1, const PoseData& P2 ) {
	EasyPoseData r( P1.source.c_str(), P2.target.c_str(), false );

	// handle covariance (the ugly part)
	TNT::Array2D<double> C( 6, 6, r.covariance[ 0 ] );
	const TNT::Array2D<double> C1( 6, 6, const_cast<double*>( P1.covariance[ 0 ] ) );
	const TNT::Array2D<double> C2( 6, 6, const_cast<double*>( P2.covariance[ 0 ] ) );

	TNT::Array2D<double> J1, J2;
	generateProductTransformMatrices( P1, P2, J1, J2 );

	TNT::Array2D<double> T1 = matmult( matmult( J1, C1 ), transpose( J1 ) );
	TNT::Array2D<double> T2 = matmult( matmult( J2, C2 ), transpose( J2 ) );

	C.inject( T1 + T2 );
	// make sure the covariance matrix stays symmetric
	normalizeSymmetric( C );

	// handle position
	if ( P1.hasPosition && P2.hasPosition ) {
		r.hasPosition = true;
		// newTrans = trans1 + (quat1 * trans2 * quat1^1)
		if ( P1.hasOrientation ) {
			double tmpPos[ 3 ];
			Util::rotateQuaternion( tmpPos, P1.orientation, P2.position );
			for ( int i = 0; i < 3; i++ )
				r.position[ i ] = tmpPos[ i ] + P1.position[ i ];

		} else
			for ( int i = 0; i < 3; i++ )
				r.position[ i ] = P1.position[ i ] + P2.position[ i ];
	}

	// handle rotation
	if ( P1.hasOrientation && P2.hasOrientation ) {
		r.hasOrientation = true;
		// newQuat = quat1*quat2
		double tmpQuat[ 4 ];
		Util::multiplyQuaternion( tmpQuat, P1.orientation, P2.orientation );
		r.setOrientation( tmpQuat );
	}

	// handle the other values
	if ( P1.timestamp < P2.timestamp )
		r.timestamp = P2.timestamp;
	else
		r.timestamp = P1.timestamp;

	r.confidence = P1.confidence * P2.confidence;
	r.timeError = P1.timeError + P2.timeError; // !?

	return r;
}


bool EasyPoseData::isSane( const PoseData& P ) {
  if ( !P.hasPosition && !P.hasOrientation )
    return false;

  if ( P.hasPosition ) {
    for ( int i = 0; i < 3; i++ )
#ifndef _MSC_VER // Win32 Hack
		if ( std::isnan( P.position[ i ] ) )
        return false;
#else
		if ( _isnan( P.position[ i ] ) )
        return false;
#endif
  }

  if ( P.hasOrientation ) {
    double d = 0.0;
    for ( int i = 0; i < 4; i++ )
      d += P.orientation[ i ] * P.orientation[ i ];
#ifndef _MSC_VER // Win32 Hack
    if ( std::isnan( d ) || d < 0.9 || d > 1.1 )
      return false;
#else
    if ( _isnan( d ) || d < 0.9 || d > 1.1 )
        return false;
#endif
  }

  return true;
}


static TNT::Array2D<double> generateInverseTransformMatrix( const PoseData& P ) {
  TNT::Array2D<double> M( 6, 6 );

  // the error (et, er) is transformed as follows (q^-1 = conjungation):
  // etn = r^-1 * t * r - er^-1 * r^-1 * ( t + et ) * r * er
  // ern = r * er^-1 * r^-1
  // now we build the jacobi matrix of this transformation evaluated at (t,r):
  // for details see Daniel's thesis

  // Maple-generated C code
  double t1 = P.orientation[0]*P.orientation[0];
  double t2 = P.orientation[3]*P.orientation[3];
  double t3 = P.orientation[2]*P.orientation[2];
  double t4 = P.orientation[1]*P.orientation[1];
  double t5 = -t1-t2+t3+t4;
  double t6 = P.orientation[1]*P.orientation[0];
  double t7 = P.orientation[2]*P.orientation[3];
  double t8 = -t6-t7;
  double t9 = P.orientation[2]*P.orientation[0];
  double t10 = P.orientation[1]*P.orientation[3];
  double t11 = -t9+t10;
  double t15 = P.orientation[0]*P.position[0]+P.orientation[1]*P.position[1]+P.orientation[2]*P.position[2];
  double t16 = t15*P.orientation[3];
  double t17 = -t15*P.orientation[3];
  double t21 = P.orientation[3]*P.position[1]+P.orientation[0]*P.position[2]-P.orientation[2]*P.position[0];
  double t25 = t21*P.orientation[0];
  double t27 = t15*P.orientation[2];
  double t31 = -P.orientation[1]*P.position[0]-P.orientation[3]*P.position[2]+P.orientation[0]*P.position[1];
  double t36 = P.orientation[3]*P.position[0]-P.orientation[1]*P.position[2]+P.orientation[2]*P.position[1];
  double t38 = 2.0*t36*P.orientation[1];
  double t39 = -t15*P.orientation[2];
  double t40 = -t31*P.orientation[3];
  double t43 = -2.0*t31*P.orientation[0];
  double t44 = t15*P.orientation[1];
  double t45 = t21*P.orientation[3];
  double t47 = t36*P.orientation[2];
  double t49 = -t15*P.orientation[1];
  double t51 = -t6+t7;
  double t52 = -t4+t3-t2+t1;
  double t53 = P.orientation[2]*P.orientation[1];
  double t54 = P.orientation[0]*P.orientation[3];
  double t55 = -t53-t54;
  double t62 = -t31*P.orientation[1];
  double t64 = t15*P.orientation[0];
  double t66 = 2.0*t21*P.orientation[2];
  double t67 = t36*P.orientation[3];
  double t69 = -t15*P.orientation[0];
  double t71 = -t9-t10;
  double t72 = -t53+t54;
  double t73 = -t3+t4+t1-t2;
  M[0][0] = t5;
  M[0][1] = 2.0*t8;
  M[0][2] = 2.0*t11;
  M[0][3] = -t16-t17;
  M[0][4] = -2.0*t25+t27-t31*P.orientation[3]+t38-t39+t40;
  M[0][5] = -t43-t44-2.0*t45+t47+t36*P.orientation[2]+t49;
  M[1][0] = 2.0*t51;
  M[1][1] = t52;
  M[1][2] = 2.0*t55;
  M[1][3] = -t38-t27+t39-2.0*t40+t25+t21*P.orientation[0];
  M[1][4] = -t16-t17;
  M[1][5] = -2.0*t62+t64+t66+t67+t36*P.orientation[3]-t69;
  M[2][0] = 2.0*t71;
  M[2][1] = 2.0*t72;
  M[2][2] = t73;
  M[2][3] = -2.0*t47+t44-t49+t43+t45+t21*P.orientation[3];
  M[2][4] = -t66-t64-t31*P.orientation[1]-2.0*t67+t69+t62;
  M[2][5] = -t16-t17;
  M[3][0] = 0.0;
  M[3][1] = 0.0;
  M[3][2] = 0.0;
  M[3][3] = t5;
  M[3][4] = 2.0*t51;
  M[3][5] = 2.0*t71;
  M[4][0] = 0.0;
  M[4][1] = 0.0;
  M[4][2] = 0.0;
  M[4][3] = 2.0*t8;
  M[4][4] = t52;
  M[4][5] = 2.0*t72;
  M[5][0] = 0.0;
  M[5][1] = 0.0;
  M[5][2] = 0.0;
  M[5][3] = 2.0*t11;
  M[5][4] = 2.0*t55;
  M[5][5] = t73;

  return M;
}


static void generateProductTransformMatrices( const PoseData& P1, const PoseData& P2,
	TNT::Array2D<double>& J1, TNT::Array2D<double>& J2 ) {
	// FIXME: may not work if P1/P2 only contains position or orientation

	// generates the two jacobi matrices necessary for error propagation by inference
	// for details see Daniel's thesis
	J1 = TNT::Array2D<double>( 6, 6 );
	J2 = TNT::Array2D<double>( 6, 6 );

	// Maple-generated!
	double t1 = P1.orientation[1]*P2.position[2];
	double t2 = P1.orientation[2]*P2.position[1];
	double t3 = t1-t2;
	double t5 = P1.orientation[1]*P2.position[1];
	double t6 = P1.orientation[2]*P2.position[2];
	double t7 = t5+t6;
	double t9 = P1.orientation[3]*P2.position[2];
	double t10 = P1.orientation[0]*P2.position[1];
	double t11 = -t9-t10;
	double t13 = P1.orientation[3]*P2.position[1];
	double t14 = P1.orientation[0]*P2.position[2];
	double t15 = t13-t14;
	double t18 = P1.orientation[2]*P2.position[0];
	double t19 = -t14+t18;
	double t21 = P1.orientation[1]*P2.position[0];
	double t22 = t9-t21;
	double t24 = P1.orientation[0]*P2.position[0];
	double t25 = t24+t6;
	double t27 = P1.orientation[3]*P2.position[0];
	double t28 = -t27-t1;
	double t31 = t10-t21;
	double t33 = -t13-t18;
	double t35 = t27-t2;
	double t37 = t24+t5;
	double t40 = P1.orientation[0]*P1.orientation[0];
	double t41 = P1.orientation[3]*P1.orientation[3];
	double t42 = P1.orientation[2]*P1.orientation[2];
	double t43 = P1.orientation[1]*P1.orientation[1];
	double t45 = P1.orientation[1]*P1.orientation[0];
	double t46 = P1.orientation[2]*P1.orientation[3];
	double t48 = P1.orientation[0]*P1.orientation[2];
	double t49 = P1.orientation[3]*P1.orientation[1];
	double t68 = P1.orientation[1]*P1.orientation[2];
	double t69 = P1.orientation[3]*P1.orientation[0];
	double t89 = P2.orientation[0]*P2.orientation[0];
	double t90 = P2.orientation[3]*P2.orientation[3];
	double t91 = P2.orientation[2]*P2.orientation[2];
	double t92 = P2.orientation[1]*P2.orientation[1];
	double t94 = P2.orientation[0]*P2.orientation[1];
	double t95 = P2.orientation[3]*P2.orientation[2];
	double t97 = P2.orientation[0]*P2.orientation[2];
	double t98 = P2.orientation[3]*P2.orientation[1];
	double t100 = t90+t89+t92+t91;
	double t103 = P2.orientation[2]*P2.orientation[1];
	double t104 = P2.orientation[0]*P2.orientation[3];

	// J1 is for P1
	J1[0][0] = 1.0;
	J1[0][1] = 0.0;
	J1[0][2] = 0.0;
	J1[0][3] = -2.0*t3*P1.orientation[0]+2.0*t7*P1.orientation[3]-2.0*t11*P1.orientation[2]+2.0*t15*P1.orientation[1];
	J1[0][4] = -2.0*t19*P1.orientation[0]+2.0*t22*P1.orientation[3]-2.0*t25*P1.orientation[2]+2.0*t28*P1.orientation[1];
	J1[0][5] = -2.0*t31*P1.orientation[0]+2.0*t33*P1.orientation[3]-2.0*t35*P1.orientation[2]+2.0*t37*P1.orientation[1];
	J1[1][0] = 0.0;
	J1[1][1] = 1.0;
	J1[1][2] = 0.0;
	J1[1][3] = -2.0*t3*P1.orientation[1]+2.0*t7*P1.orientation[2]+2.0*t11*P1.orientation[3]-2.0*t15*P1.orientation[0];
	J1[1][4] = -2.0*t19*P1.orientation[1]+2.0*t22*P1.orientation[2]+2.0*t25*P1.orientation[3]-2.0*t28*P1.orientation[0];
	J1[1][5] = -2.0*t31*P1.orientation[1]+2.0*t33*P1.orientation[2]+2.0*t35*P1.orientation[3]-2.0*t37*P1.orientation[0];
	J1[2][0] = 0.0;
	J1[2][1] = 0.0;
	J1[2][2] = 1.0;
	J1[2][3] = -2.0*t3*P1.orientation[2]-2.0*t7*P1.orientation[1]+2.0*t11*P1.orientation[0]+2.0*t15*P1.orientation[3];
	J1[2][4] = -2.0*t19*P1.orientation[2]-2.0*t22*P1.orientation[1]+2.0*t25*P1.orientation[0]+2.0*t28*P1.orientation[3];
	J1[2][5] = -2.0*t31*P1.orientation[2]-2.0*t33*P1.orientation[1]+2.0*t35*P1.orientation[0]+2.0*t37*P1.orientation[3];
	J1[3][0] = 0.0;
	J1[3][1] = 0.0;
	J1[3][2] = 0.0;
	J1[3][3] = t89+t90-t91-t92;
	J1[3][4] = 2.0*t94+2.0*t95;
	J1[3][5] = 2.0*t97-2.0*t98;
	J1[4][0] = 0.0;
	J1[4][1] = 0.0;
	J1[4][2] = 0.0;
	J1[4][3] = 2.0*t94-2.0*t95;
	J1[4][4] = t92-t91+t90-t89;
	J1[4][5] = 2.0*t103+2.0*t104;
	J1[5][0] = 0.0;
	J1[5][1] = 0.0;
	J1[5][2] = 0.0;
	J1[5][3] = 2.0*t97+2.0*t98;
	J1[5][4] = 2.0*t103-2.0*t104;
	J1[5][5] = t91-t92-t89+t90;

	// J2 is for P2
	// this one could be decomposed into two blocks for better performance
	J2[0][0] = t40+t41-t42-t43;
	J2[0][1] = 2.0*t45-2.0*t46;
	J2[0][2] = 2.0*t48+2.0*t49;
	J2[0][3] = 0.0;
	J2[0][4] = 0.0;
	J2[0][5] = 0.0;
	J2[1][0] = 2.0*t45+2.0*t46;
	J2[1][1] = t43-t42+t41-t40;
	J2[1][2] = 2.0*t68-2.0*t69;
	J2[1][3] = 0.0;
	J2[1][4] = 0.0;
	J2[1][5] = 0.0;
	J2[2][0] = 2.0*t48-2.0*t49;
	J2[2][1] = 2.0*t68+2.0*t69;
	J2[2][2] = t42-t43-t40+t41;
	J2[2][3] = 0.0;
	J2[2][4] = 0.0;
	J2[2][5] = 0.0;
	J2[3][0] = 0.0;
	J2[3][1] = 0.0;
	J2[3][2] = 0.0;
	J2[3][3] = t100;
	J2[3][4] = 0.0;
	J2[3][5] = 0.0;
	J2[4][0] = 0.0;
	J2[4][1] = 0.0;
	J2[4][2] = 0.0;
	J2[4][3] = 0.0;
	J2[4][4] = t100;
	J2[4][5] = 0.0;
	J2[5][0] = 0.0;
	J2[5][1] = 0.0;
	J2[5][2] = 0.0;
	J2[5][3] = 0.0;
	J2[5][4] = 0.0;
	J2[5][5] = t100;
	}


//Code by Daniel Pustka: see src/services/ubitrack/PoseTool

std::ostream& DWARF::operator<<( std::ostream& os, const PoseData& pose )
   {
   os << pose.position[0] << " " << pose.position[1] << " " << pose.position[2] << " ";
   os << pose.orientation[0] << " " << pose.orientation[1] << " " << pose.orientation[2] << " " << pose.orientation[3] << " ";
   os << pose.timestamp.seconds << "." << std::setw( 6 ) << std::setfill( '0' ) << pose.timestamp.microseconds << " ";

  // additional fields added for the new ubitrack posedata
   os << pose.hasPosition << " " << pose.hasOrientation << " " << pose.confidence << " ";
   for ( int i = 0; i < 6; i++ )
      for ( int j = 0; j < 6; j++ )
         os << pose.covariance[i][j] << " ";
   os << pose.timeError;
   os << std::endl;

   return os;
   }


std::istream& DWARF::operator>>( std::istream& is, PoseData& Pose )
  {
  is >> Pose.position[0] >> Pose.position[1] >> Pose.position[2];
  is >> Pose.orientation[0] >> Pose.orientation[1] >> Pose.orientation[2] >> Pose.orientation[3];
  is >> Pose.timestamp.seconds;
  is.ignore() >> Pose.timestamp.microseconds;

  // additional fields added for the new ubitrack posedata
  is >> Pose.hasPosition >> Pose.hasOrientation >> Pose.confidence;
  for ( int i = 0; i < 6; i++ )
    for ( int j = 0; j < 6; j++ )
      is >> Pose.covariance[i][j];
  is >> Pose.timeError;

  //DEBUGSTREAM( 50, "Loaded PoseData: t=" << Pose.timestamp.seconds << "." << Pose.timestamp.microseconds <<
  // ", p=" << Pose.position[ 0 ] << ", " << Pose.position[ 1 ] << ", ..." );

  return is;
  }

