/***************************************************************************
 *                         EasyPoseData.h  -  description
 *                            -------------------
 *   begin                : 2004/03/31
 *   copyright            : (C) 2004 by Daniel Pustka
 *   email                : pustka@in.tum.de
 *
 *   $Id: EasyPoseData.h,v 1.6 2004/10/19 17:52:11 strassef Exp $
 *   $Revision: 1.6 $
 *
  ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef __EasyPoseData_H__
#define __EasyPoseData_H__

#include <string>
#include <iostream>
#include <string.h>
#include <dwarfutil.h>
#include <DWARF/PoseData.h>

namespace DWARF {

  /**
   * @brief A wrapper for the PoseData structure with some convenience
   */
  class EasyPoseData : public DWARF::PoseData {

  public:

    /**
     * @brief Default constructor
     * Initializes all fields to some default values
     * @param pSource Value for the source object ID
     * @param pTarget Value for the target object ID
     * @param bInitializeError should covariance be initialized with default values?
     */
    EasyPoseData( const char* pSource = NULL, const char* pTarget = NULL,
      bool bInitializeError = true );

    /**
     * @brief Copies all values from a PoseData
     */
    EasyPoseData( const DWARF::PoseData& p )
      { *static_cast<DWARF::PoseData*>( this ) = p; }

    /**
     * @brief Copies all values from a PoseData
     */
    EasyPoseData& operator=( const DWARF::PoseData& p )
      { *static_cast<DWARF::PoseData*>( this ) = p; return *this; }


    // plenty of member access functions

    /** set the source ID */
    void setSource( const char* s )
      { source = s; }

    /** set the source ID */
    void setSource( const std::string& s )
      { setSource( s.c_str() ); }

    /** set the source ID */
    void setSource( const std::string* s )
      { setSource( s->c_str() ); }

    /** get the source ID */
    const char* getSource() const
      { return source.c_str(); }


    /** set the target ID */
    void setTarget( const char* t )
      { target = t; }

    /** set the target ID */
    void setTarget( const std::string& t )
      { setTarget( t.c_str() ); }

    /** set the target ID */
    void setTarget( const std::string* t )
      { setTarget( t->c_str() ); }

    /** get the target ID */
    const char* getTarget() const
      { return target.c_str(); }


    /** set single position value */
    void setPosition( int index, double value )
      { hasPosition = true; position[ index ] = value; }

    /** set whole position from array */
    void setPosition( const double* p )
      { hasPosition = true; memcpy( position, p, sizeof( double[ 3 ] ) ); }

    /** get single position value */
    double getPosition( int index ) const
      { return position[ index ]; }

    /** get whole position */
    void getPosition( double* p ) const
      { memcpy( p, position, sizeof( double[ 3 ] ) ); }


    /** set single orientation value */
    void setOrientation( int index, double value )
      { hasOrientation = true; orientation[ index ] = value; }

    /** set whole position from array */
    void setOrientation( const double* o )
      { hasOrientation = true; memcpy( orientation, o, sizeof( double[ 4 ] ) ); }

    /** get single orientation value */
    double getOrientation( int index ) const
      { return orientation[ index ]; }

    /** get whole orientation */
    void getOrientation( double* o ) const
      { memcpy( o, orientation, sizeof( double[ 4 ] ) ); }


    /** set the timestamp */
    void setTimestamp( const DWARF::Time* t )
      { timestamp = *t; }

    /** set the timestamp */
    void setTimestamp( const DWARF::Time& t )
      { timestamp = t; }

    /** get the timestamp */
    Time getTimestamp() const
      { return timestamp; }

    /** set the timestamp to the current time */
    void updateTimestamp()
      { timestamp = DWARF::Util::currentTime(); }


    /** set the measurement confidence  */
    void setConfidence( double c )
      { confidence = c; }

    /** get the measurement confidence  */
    double getConfidence() const
      { return confidence; }


    /**
     * sets the position part of the covariance matrix to I*sigma^2
     * This means that all variables are assumed to be independent
     * @ param sigma standard deviation
     */
    void setPositionAccuracy( double sigma );

    /**
     * @brief Sets the orientation part of the covariance matrix to I*sigma^2
     * This means that all variables are assumed to be independent
     * @ param sigma standard deviation
     */
    void setOrientationAccuracy( double sigma );


    /** sets the time error (standard deviation) */
    void setTimeError( double sigma )
      { timeError = sigma; }

    /** gets the time error (standard deviation) */
    double getTimeError() const
      { return timeError; }


    /**
     * returns an inverted PoseData with correctly transformed covariance
     * Note: doing this twice will increase the error!
     */
    static EasyPoseData inverse( const PoseData& P );

    /** inverse( *this ) */
    EasyPoseData inverse() const
      { return inverse( *this ); }


    /**
     * computes the product of two PoseData structures with correctly transformed covariance
     */
    static EasyPoseData product( const PoseData& P1, const PoseData& P2 );

    /** product( *this, P ) */
    EasyPoseData product( const PoseData& P )
      { return product( *this, P ); }


    /**
     * pose sanity checks
     * checks for NANs, quaternion length, etc
     */
    static bool isSane( const PoseData& P );

    /** isSane( *this ) */
    bool isSane() const
      { return isSane( *this ); }

  };

  /** Ostream operator for PoseData */
  std::ostream& operator<<(std::ostream& os, const PoseData& p);

  /** Istream operator for PoseData */
  std::istream& operator>>(std::istream& os, PoseData& p);

};

#endif
