/* ========================================================================
 * Copyright (C) 2000-2004  Technische Universitaet Muenchen
 * License: GPL v2 (or later)
 *
 * For further information please contact the DWARF team at
 * <dwarf-gnu@augmentedreality.de>
 *
 * DerivThread.h
 * Daniel Pustka <pustka@in.tum.de> May 2004
 * Distributed Wearable Augmented Reality Framework - www.augmentedreality.de
 * For documentation and instructions, see the DWARF documentation website at
 * http://wwwbruegge.in.tum.de/projects/dwarf/doc/
 *
 * For further questions about DWARF, please contact us at
 * <dwarf@augmentedreality.de>
 * For questions about this particular service, contact
 * Daniel Pustka <pustka@in.tum.de>
 *
 * $Id: DerivThread.cpp,v 1.1 2004/08/14 23:32:39 pustka Exp $
 */

#include "DerivThread.h"
#include <debug.h>
#include <TimeUtil.h>
#include <qapplication.h>
#include <algorithm>
#include <ubitrack/EasyPoseData.h>

using DWARF::Time;
using DWARF::PoseData;
using DWARF::EasyPoseData;
using std::vector;


DerivThread::DerivThread( PoseSequence& Destination, const PoseSequenceSubCollection& Sources,
	int nDerivation, QObject* pReceiver, QEvent* pEvent )
	: m_Destination( Destination )
	, m_Sequences( Sources )
	, m_pReceiver( pReceiver )
	, m_pEvent( pEvent )
	, m_nDerivation( nDerivation )
	, m_bQuit( false )
	{
	}


void DerivThread::run()
	{
	DEBUGSTREAM( 20, "DerivThread started" );

	m_Sequences[ 0 ]->GetMutex().lock();

	// erase destination
	m_Destination.GetMutex().lock();
	m_Destination.clear();
	m_Destination.SetDerivations( m_nDerivation, m_nDerivation );

	DEBUGSTREAM( 30, "DerivThread locked" );

	vector<PoseData> DerivBuffer( m_nDerivation + 1 );

	int cEvent = 0;
	for ( PoseSequence::SequenceType::const_iterator it = m_Sequences[ 0 ]->begin();
		it != m_Sequences[ 0 ]->end() && !m_bQuit; it++ )
		{
		if ( it->timestamp <= DerivBuffer[ 0 ].timestamp )
			continue;

		Time DeltaTime = it->timestamp - DerivBuffer[ 0 ].timestamp;
		double fDeltaTime = DeltaTime.seconds + DeltaTime.microseconds / 1000000.0;
		DerivBuffer[ 0 ].timestamp = it->timestamp;

		EasyPoseData DiffResult( *it );

		// differentiate position
		if ( DiffResult.hasPosition )
			for ( int d = 0; d < m_nDerivation; d++ )
				for ( int i = 0; i < 3; i++ )
					{
					double f = ( DiffResult.position[ i ] - DerivBuffer[ d ].position[ i ] ) / fDeltaTime;
					DerivBuffer[ d ].position[ i ] = DiffResult.position[ i ];
					DiffResult.position[ i ] = f;
					}

		// differentiate orientation
		if ( DiffResult.hasOrientation )
			{
			// first derivation
			if ( m_nDerivation )
				{
				// compute difference quaternion
				double q[ 4 ];
				DWARF::Util::quaternionNegateIfCloser( DiffResult.orientation, DerivBuffer[ 0 ].orientation );
				DWARF::Util::invertQuaternion( DerivBuffer[ 0 ].orientation );
				DWARF::Util::multiplyQuaternion( q, DerivBuffer[ 0 ].orientation, DiffResult.orientation );

				for ( int i = 0; i < 4; i++ )
					DerivBuffer[ 0 ].orientation[ i ] = DiffResult.orientation[ i ];

				double fAngSpeed;
				if ( q[ 3 ] >= 1.0 || q[ 3 ] <= -1.0 )
					fAngSpeed = 0.0;
				else
					if ( q[ 3 ] >= 0 )
						fAngSpeed = 2 * acos( q[ 3 ] ) / fDeltaTime;
					else
						fAngSpeed = 2 * -acos( -q[ 3 ] ) / fDeltaTime;

				double qFact = 0.0;
				for ( int i = 0; i < 3; i++ )
					qFact += q[ i ] * q[ i ];
				if ( qFact > 10e-10 )
					qFact = fAngSpeed / sqrt( qFact );
				else
					qFact = 0.0;

				for ( int i = 0; i < 3; i++ )
					DiffResult.orientation[ i ] = q[ i ] * qFact;
				DiffResult.orientation[ 3 ] = fAngSpeed; // 0.0;

				DEBUGSTREAM( 26, "AngSpeed = " << fAngSpeed << ", qFact = " << qFact <<
					", DiffQ = [ " << q[0] << ", " << q[1] << ", " << q[2] << ", " << q[3] << " ]" );
				DEBUGSTREAM( 25, "AngVel = [ " <<
					DiffResult.orientation[0] << ", " << DiffResult.orientation[1] << ", " <<
					DiffResult.orientation[2] << ", 0 ]" );
				}

			// linear for higher derivations
			for ( int d = 1; d < m_nDerivation; d++ )
				for ( int i = 0; i < 3; i++ )
					{
					double f = ( DiffResult.orientation[ i ] - DerivBuffer[ d ].orientation[ i ] ) / fDeltaTime;
					DerivBuffer[ d ].orientation[ i ] = DiffResult.orientation[ i ];
					DiffResult.orientation[ i ] = f;
					}
			}

		if ( cEvent >= m_nDerivation )
			m_Destination.push_back( DiffResult );
		cEvent++;
		}

	// unlock sequences
	m_Destination.GetMutex().unlock();
	m_Sequences[ 0 ]->GetMutex().unlock();

	// send event only when finished completely
	if ( !m_bQuit )
		{
		DEBUGSTREAM( 20, "DerivThread finished normally" );
		QApplication::postEvent( m_pReceiver, m_pEvent );
		}
	else
		{
		DEBUGSTREAM( 20, "DerivThread stopped unfinished" );
		delete m_pEvent;
		}
	}

