/* ========================================================================
 * Copyright (C) 2000-2004  Technische Universitaet Muenchen
 * License: GPL v2 (or later)
 *
 * For further information please contact the DWARF team at
 * <dwarf-gnu@augmentedreality.de>
 *
 * EkfWindowImpl.h
 * Daniel Pustka <pustka@in.tum.de> May 2004
 * Distributed Wearable Augmented Reality Framework - www.augmentedreality.de
 * For documentation and instructions, see the DWARF documentation website at
 * http://wwwbruegge.in.tum.de/projects/dwarf/doc/
 *
 * For further questions about DWARF, please contact us at
 * <dwarf@augmentedreality.de>
 * For questions about this particular service, contact
 * Daniel Pustka <pustka@in.tum.de>
 *
 * $Id: EkfThread.cpp,v 1.2 2004/08/14 23:32:39 pustka Exp $
 */

#include "EkfThread.h"
#include "PoseSequenceView.h"
#include <qslider.h>
#include <qcheckbox.h>
#include <qlineedit.h>
#include <qspinbox.h>
#include <qgroupbox.h>
#include <debug.h>
#include <dwarfutil.h>
#include <TimeUtil.h>
#include <qapplication.h>
#include <algorithm>

using DWARF::Time;
using DWARF::PoseData;
using std::vector;


EkfThread::EkfThread( PoseSequence& Destination, const PoseSequenceSubCollection& Sources,
   const PoseEKF::MotionModel& MoMo, int nPredict, int nSampleInterval,
   QObject* pReceiver, QEvent* pEvent )
   : m_Destination( Destination )
   , m_Sequences( Sources )
   , m_MoMo( MoMo)
   , m_nPredict( nPredict )
   , m_nSampleInterval( nSampleInterval )
   , m_pReceiver( pReceiver )
   , m_pEvent( pEvent )
   , m_bQuit( false )
   {
   }


void EkfThread::run()
   {
   DEBUGSTREAM( 20, "EkfThread started" );
   PoseEKF EKF( m_MoMo );

   // compute a predicted state every 10 ms
   Time StepSize = { m_nSampleInterval / 1000, ( m_nSampleInterval % 1000 ) * 1000 };
   Time PredictTime = { std::abs( m_nPredict ) / 1000, ( std::abs( m_nPredict ) % 1000 ) * 1000 };

   // vector with iterator for each sequence
   vector<PoseSequence::SequenceType::const_iterator> Its;

   // lock all sequences and push iterator into vector
   for ( unsigned i = 0; i < m_Sequences.size(); i++ )
      {
      m_Sequences[ i ]->GetMutex().lock();
      Its.push_back( m_Sequences[ i ]->begin() );
      }

   // erase destination
   m_Destination.GetMutex().lock();
   m_Destination.clear();

   DEBUGSTREAM( 30, "EkfThread locked" );

   // find first event
   Time ItTime = Its[ 0 ]->timestamp;
   int iIt = 0;
   for ( unsigned i = 1; i < Its.size(); i++ )
      if ( Its[ i ]->timestamp < ItTime )
         ItTime = Its[ i ]->timestamp, iIt = i;

	// align output times to multiples of StepSize
	Time AlignTmpTime = m_nPredict < 0 ? ( Its[ iIt ]->timestamp - PredictTime ) :
		( Its[ iIt ]->timestamp + PredictTime );
	unsigned nAlignBase = AlignTmpTime.seconds - ( AlignTmpTime.seconds % m_nSampleInterval );
	AlignTmpTime.seconds -= nAlignBase;
	unsigned nAlignMs = AlignTmpTime.seconds * 1000 + AlignTmpTime.microseconds / 1000;
	nAlignMs = nAlignMs - ( nAlignMs % m_nSampleInterval ) + m_nSampleInterval;
	AlignTmpTime.seconds = nAlignBase + nAlignMs / 1000;
	AlignTmpTime.microseconds = ( nAlignMs % 1000 ) * 1000;

	// compute measurement time of first output
	Time NextPredict = m_nPredict < 0 ? ( AlignTmpTime + PredictTime ) :
		( AlignTmpTime - PredictTime );

	// variables for performance measurements
	int cPredictions = 0;
	int cMeasurements = 0;
	Time PerfMonTime = DWARF::Util::currentTime();

   while ( !m_bQuit )
      {
      if ( NextPredict < ItTime )
         {
         // predict
			cPredictions++;
         PoseData D = EKF.PredictPose( m_nPredict < 0 ? ( NextPredict - PredictTime ) :
            ( NextPredict + PredictTime ) );
         m_Destination.push_back( D );

         NextPredict = NextPredict + StepSize;
         }
      else
         {
         // measure
			cMeasurements++;
         EKF.AddPose( *Its[ iIt ], m_Sequences[ iIt ]->GetPosDerivation(),
				m_Sequences[ iIt ]->GetOriDerivation() );

         // advance file
         Its[ iIt ]++;

         // find next measurement
         ItTime.seconds = 0x7FFFFFFF;
         iIt = -1;
         for ( unsigned i = 0; i < Its.size(); i++ )
            if ( Its[ i ] != m_Sequences[ i ]->end() && Its[ i ]->timestamp < ItTime )
               ItTime = Its[ i ]->timestamp, iIt = i;

         if ( iIt < 0 )
            break;
         }
      }

	PerfMonTime = DWARF::Util::currentTime() - PerfMonTime;

	// unlock sequences
	m_Destination.GetMutex().unlock();
	for ( unsigned i = 0; i < m_Sequences.size(); i++ )
		m_Sequences[ i ]->GetMutex().unlock();

	// send event only when finished completely
	if ( iIt < 0 )
		{
		DEBUGSTREAM( 20, "EkfThread finished normally" );
		QApplication::postEvent( m_pReceiver, m_pEvent );

		if ( cPredictions > 0 && cMeasurements > 0 )
			{
			double fPerfMonTime = PerfMonTime.seconds + PerfMonTime.microseconds / 1000000.0;
			DEBUGSTREAM( 11, "EKF Measurements/s: " << cMeasurements / fPerfMonTime );
			DEBUGSTREAM( 11, "EKF Predictions/s:  " << cPredictions / fPerfMonTime );
			}
		}
	else
		{
		DEBUGSTREAM( 20, "EkfThread stopped unfinished" );
		delete m_pEvent;
		}
	}

