/* ========================================================================
 * Copyright (C) 2000-2004  Technische Universitaet Muenchen
 * License: GPL v2 (or later)
 *
 * For further information please contact the DWARF team at
 * <dwarf-gnu@augmentedreality.de>
 *
 * EkfWindowImpl.h
 * Daniel Pustka <pustka@in.tum.de> May 2004
 * Distributed Wearable Augmented Reality Framework - www.augmentedreality.de
 * For documentation and instructions, see the DWARF documentation website at
 * http://wwwbruegge.in.tum.de/projects/dwarf/doc/
 *
 * For further questions about DWARF, please contact us at
 * <dwarf@augmentedreality.de>
 * For questions about this particular service, contact
 * Daniel Pustka <pustka@in.tum.de>
 *
 * $Id: EkfWindowImpl.cpp,v 1.2 2004/08/14 23:32:39 pustka Exp $
 */

#include "EkfWindowImpl.h"
#include "PoseSequenceView.h"
#include <qslider.h>
#include <qcheckbox.h>
#include <qlineedit.h>
#include <qspinbox.h>
#include <qgroupbox.h>
#include <qmessagebox.h>
#include <debug.h>
#include <qpushbutton.h>
#include "AutoTuner.h"
#include <config.h>

static const double g_fLn10 = 2.30258509299404568401; // ln( 10 ) - used to compute 10^x from e^x

class EkfThreadFinishedEvent : public QEvent
   {
   public:
      static const int g_nEvID = QEvent::User + 164;
      int m_nThreadID;

      EkfThreadFinishedEvent( int nThreadID )
         : QEvent( static_cast<QEvent::Type>( g_nEvID ) )
         , m_nThreadID( nThreadID )
         {}

      virtual ~EkfThreadFinishedEvent() {}
   };



EkfWindowImpl::EkfWindowImpl( QWidget* pParent, const char* pszName )
   : EKFWindow( pParent, pszName )
   , m_pDestinationSequence( 0 )
   , m_nThreadID( 0 )
   {
   setWFlags( Qt::WDestructiveClose );

   // no way to find out which control sent a signal in Qt
   m_Controls.push_back( EditPair( SliderPNPosX0, EditPNPosX0 ) );
   m_Controls.push_back( EditPair( SliderPNPosY0, EditPNPosY0 ) );
   m_Controls.push_back( EditPair( SliderPNPosZ0, EditPNPosZ0 ) );
   m_Controls.push_back( EditPair( SliderPNPosX1, EditPNPosX1 ) );
   m_Controls.push_back( EditPair( SliderPNPosY1, EditPNPosY1 ) );
   m_Controls.push_back( EditPair( SliderPNPosZ1, EditPNPosZ1 ) );
   m_Controls.push_back( EditPair( SliderPNPosX2, EditPNPosX2 ) );
   m_Controls.push_back( EditPair( SliderPNPosY2, EditPNPosY2 ) );
   m_Controls.push_back( EditPair( SliderPNPosZ2, EditPNPosZ2 ) );
   m_Controls.push_back( EditPair( SliderPNPosX3, EditPNPosX3 ) );
   m_Controls.push_back( EditPair( SliderPNPosY3, EditPNPosY3 ) );
   m_Controls.push_back( EditPair( SliderPNPosZ3, EditPNPosZ3 ) );
   m_Controls.push_back( EditPair( SliderPNOri0, EditPNOri0 ) );
   m_Controls.push_back( EditPair( SliderPNOri1, EditPNOri1 ) );
   m_Controls.push_back( EditPair( SliderPNOri2, EditPNOri2 ) );
   m_Controls.push_back( EditPair( SliderPNOri3, EditPNOri3 ) );

   // connect controls
   for ( unsigned i = 0; i < m_Controls.size(); i++ )
      {
      connect( m_Controls[ i ].pSlider, SIGNAL( sliderMoved(int) ),
         this, SLOT( SliderChanged() ) );
      connect( m_Controls[ i ].pEdit, SIGNAL( returnPressed() ),
         this, SLOT( ReturnPressed() ) );
      }

   UpdateEnabled();
   
//   #ifndef HAVE_GSL
   ButtonAutoTune->setEnabled( false );
//   #endif
   }


EkfWindowImpl::~EkfWindowImpl()
   {
   // unreference sequences
   StopThread();
   SetSequences( PoseSequenceSubCollection() );
   }


void EkfWindowImpl::SetSequences( const PoseSequenceSubCollection& Seqs )
   {
   // disconnect and unreference old sequences
   for ( unsigned i = 0; i < m_Sequences.size(); i++ )
      {
      disconnect( m_Sequences[ i ], SIGNAL( Changed() ), this, SLOT( SequenceChanged() ) );
      m_Sequences[ i ]->RemoveRef();
      }

   // copy sequence pointers
   m_Sequences = Seqs;

   if ( m_Sequences.empty() )
      return;

   // disconnect and reference sequences
   for ( unsigned i = 0; i < m_Sequences.size(); i++ )
      {
      connect( m_Sequences[ i ], SIGNAL( Changed() ), this, SLOT( SequenceChanged() ) );
      m_Sequences[ i ]->AddRef();
      }

   // create a new sequence for the result
   std::string sName = g_pPoseSequenceCollection->CreateNewName(
      ( "EKF-" + m_Sequences[ 0 ]->GetName() ).c_str() );
   m_pDestinationSequence = g_pPoseSequenceCollection->AddSequence( sName.c_str() );

   // adapt controls to motion model
   CheckPosition->setChecked( m_MoMo.GetPosDerivations() >= 0 );
   CheckOrientation->setChecked( m_MoMo.GetOriDerivations() >= 0 );
   SpinPosDerivations->setValue( std::max( 0, m_MoMo.GetPosDerivations() ) );
   SpinOriDerivations->setValue( std::max( 0, m_MoMo.GetOriDerivations() ) );
   UpdateEnabled();
   UpdateControls();

   StartThread();
   }


void EkfWindowImpl::UpdateEnabled()
   {
   bool bPositionEnabled = CheckPosition->isChecked();
   bool bOrientationEnabled = CheckOrientation->isChecked();
   SpinPosDerivations->setEnabled( bPositionEnabled );
   SpinOriDerivations->setEnabled( bOrientationEnabled );
   CheckXYZIndependent->setEnabled( bPositionEnabled );
   GroupPNX->setEnabled( bPositionEnabled );
   GroupPNO->setEnabled( bOrientationEnabled );

   bool bIndependent = CheckXYZIndependent->isChecked();
   GroupPNY->setEnabled( bIndependent && bPositionEnabled );
   GroupPNZ->setEnabled( bIndependent && bPositionEnabled );

   int nPosDerivations = SpinPosDerivations->value();
   for ( int i = 0; i < 12; i++ )
      {
      m_Controls[ i ].pSlider->setEnabled( i / 3 <= nPosDerivations );
      m_Controls[ i ].pEdit->setEnabled( i / 3 <= nPosDerivations );
      }

   int nOriDerivations = SpinOriDerivations->value();
   for ( int i = 0; i < 4; i++ )
      {
      m_Controls[ 12 + i ].pSlider->setEnabled( i <= nOriDerivations );
      m_Controls[ 12 + i ].pEdit->setEnabled( i <= nOriDerivations );
      }
   }


void EkfWindowImpl::UpdateControlPair( int iControl, double fVal )
   {
   char buf[10];
   sprintf( buf, "%1.1e", fVal );
   m_Controls[ iControl ].pSlider->setValue( static_cast<int>( log10( fVal ) * 20 ) );
   m_Controls[ iControl ].pEdit->setText( buf );
   }


void EkfWindowImpl::UpdateControls()
   {
   for ( int i = 0; i < m_MoMo.GetPosDerivations() + 1; i++ )
      for ( int j = 0; j < 3; j++ )
         UpdateControlPair( i * 3 + j, m_MoMo.GetPosProcessNoise( j, i ) );

   for ( int i = 0; i < m_MoMo.GetOriDerivations() + 1; i++ )
      UpdateControlPair( 12 + i, m_MoMo.GetOriProcessNoise( i ) );
   }


void EkfWindowImpl::SliderChanged()
   {
   // update position
   bool bIndependent = CheckXYZIndependent->isChecked();
   for ( int i = 0; i < m_MoMo.GetPosDerivations() + 1; i++ )
      for ( int j = 0; j < ( bIndependent ? 3 : 1 ); j++ )
         {
         double fVal = exp( m_Controls[ i * 3 + j ].pSlider->value() * g_fLn10 / 20 );
         if ( fVal != m_MoMo.GetPosProcessNoise( j, i ) )
            for ( int k = ( bIndependent ? j : 0 ); k < ( bIndependent ? j + 1 : 3 ); k++ )
               {
               m_MoMo.SetPosProcessNoise( k, i, fVal );
               UpdateControlPair( i * 3 + k, fVal );
               }
         }

   // update orientation
   for ( int i = 0; i < m_MoMo.GetOriDerivations() + 1; i++ )
      {
      double fVal = exp( m_Controls[ 12 + i ].pSlider->value() * g_fLn10 / 20 );
      if ( fVal != m_MoMo.GetOriProcessNoise( i ) )
         {
         m_MoMo.SetOriProcessNoise( i, fVal );
         UpdateControlPair( 12 + i, fVal );
         }
      }

   StartThread();
   }


void EkfWindowImpl::ReturnPressed()
   {
   // TODO: implement
   DEBUGSTREAM( 1, "Not implemented" );
   //StartThread();
   }


void EkfWindowImpl::DerivationsChanged()
   {
   m_MoMo.SetPosDerivations( CheckPosition->isChecked() ? SpinPosDerivations->value() : -1 );
   m_MoMo.SetOriDerivations( CheckOrientation->isChecked() ? SpinOriDerivations->value() : -1 );

   // copy dependent y and z
   if ( !CheckXYZIndependent->isChecked() )
      for ( int i = 0; i < m_MoMo.GetPosDerivations() + 1; i++ )
         for ( int j = 1; j < 3; j++ )
            m_MoMo.SetPosProcessNoise( j, i, m_MoMo.GetPosProcessNoise( 0, i ) );

   UpdateEnabled();
   UpdateControls();

   StartThread();
   }


void EkfWindowImpl::SomethingElseChanged()
   {
   StartThread();
   }


void EkfWindowImpl::SequenceChanged()
   {
   StartThread();
   }


bool EkfWindowImpl::event( QEvent* pEvent )
   {
   if ( pEvent->type() == EkfThreadFinishedEvent::g_nEvID )
      {
      DEBUGSTREAM( 20, "EkfWindow received finished() notification from EkfThread" );
      if ( static_cast<EkfThreadFinishedEvent*>( pEvent )->m_nThreadID == m_nThreadID )
         {
         StopThread();
         m_pDestinationSequence->SignalChanged();
         }
      else
         { DEBUGSTREAM( 20, "...from old thread" ); }
      return true;
      }
   else
      return EKFWindow::event( pEvent );
   }


void EkfWindowImpl::StartThread()
   {
   if ( m_Sequences.empty() || ( m_MoMo.GetPosDerivations() < 0 && m_MoMo.GetOriDerivations() < 0 ) )
      return;

   StopThread();
   m_pThread = std::auto_ptr<EkfThread>( new EkfThread( *m_pDestinationSequence, m_Sequences,
      m_MoMo, SpinPredict->value(), SpinStep->value(), this,
      new EkfThreadFinishedEvent( ++m_nThreadID ) ) );
   m_pThread->start();
   }


void EkfWindowImpl::StopThread()
   {
   if ( m_pThread.get() )
      {
      DEBUGSTREAM( 20, "Stopping thread" );
      m_pThread->Quit();
      m_pThread->wait();
      m_pThread.reset();
      }
   }


void EkfWindowImpl::AutoTune()
   {
#ifdef HAVE_GSL
   StopThread();
   AutoTuner AT( m_Sequences, m_MoMo, SpinPredict->value(), CheckXYZIndependent->isChecked() );
   AT.OptimizePN();
   m_MoMo = AT.GetMotionModel();
   UpdateControls();
   StartThread();
#endif   
   }


void EkfWindowImpl::PredictionError()
   {
   AutoTuner AT( m_Sequences, m_MoMo, SpinPredict->value(), CheckXYZIndependent->isChecked() );
   double fErr;
   double fRmsErr = AT.ComputePredictionError( fErr );
   fRmsErr = sqrt( fRmsErr );
   QMessageBox::information( this, "Prediction Error",
      QString().sprintf( "Average: %f\tRMS: %f", fErr, fRmsErr ) );
   }
