/* ========================================================================
 * Copyright (C) 2000-2004  Technische Universitaet Muenchen
 * License: GPL v2 (or later)
 *
 * For further information please contact the DWARF team at
 * <dwarf-gnu@augmentedreality.de>
 *
 * InfThread.h
 * Daniel Pustka <pustka@in.tum.de> May 2004
 * Distributed Wearable Augmented Reality Framework - www.augmentedreality.de
 * For documentation and instructions, see the DWARF documentation website at
 * http://wwwbruegge.in.tum.de/projects/dwarf/doc/
 *
 * For further questions about DWARF, please contact us at
 * <dwarf@augmentedreality.de>
 * For questions about this particular service, contact
 * Daniel Pustka <pustka@in.tum.de>
 *
 * $Id: InfThread.cpp,v 1.1 2004/08/14 23:32:39 pustka Exp $
 */

#include "InfThread.h"
#include <debug.h>
#include <TimeUtil.h>
#include <qapplication.h>
#include <algorithm>
#include <ubitrack/EasyPoseData.h>

using DWARF::Time;
using DWARF::PoseData;
using DWARF::EasyPoseData;
using std::vector;


InfThread::InfThread( PoseSequence& Destination, const PoseSequenceSubCollection& Sources,
   QObject* pReceiver, QEvent* pEvent )
   : m_Destination( Destination )
   , m_Sequences( Sources )
   , m_pReceiver( pReceiver )
   , m_pEvent( pEvent )
   , m_bQuit( false )
   {
   }


void InfThread::run()
   {
   DEBUGSTREAM( 20, "InfThread started" );

   // vector with iterator for each sequence
   vector<PoseSequence::SequenceType::const_iterator> Its;
   vector<EasyPoseData> PoseArray( m_Sequences.size() );

   // lock all sequences and push iterator into vector
   Time CurTime = DWARF::Util::minTime;
   for ( unsigned i = 0; i < m_Sequences.size(); i++ )
      {
      m_Sequences[ i ]->GetMutex().lock();
      Its.push_back( m_Sequences[ i ]->begin() );

      if ( m_Sequences[ i ]->begin() != m_Sequences[ i ]->end() )
         {
         // copy first event to array
         PoseArray[ i ] = m_Sequences[ i ]->front();

         // find last of all first events
         if ( CurTime < m_Sequences[ i ]->front().timestamp )
            CurTime = m_Sequences[ i ]->front().timestamp;
         }
      }

   // erase destination
   m_Destination.GetMutex().lock();
   m_Destination.clear();

   DEBUGSTREAM( 30, "InfThread locked" );

   while ( !m_bQuit && CurTime != DWARF::Util::maxTime )
      {
      // Advance all sequences to CurTime:
      // inference only works when events from all sequences are available
      for ( unsigned i = 0; i < Its.size(); i++ )
         while ( Its[ i ] != m_Sequences[ i ]->end() &&
            ( Its[ i ]->timestamp < CurTime || Its[ i ]->timestamp == CurTime ) )
            PoseArray[ i ] = *Its[ i ]++;

      // perform inference on the elements of PoseArray
		EasyPoseData P = PoseArray[ 0 ];
		for ( unsigned i = 1; i < PoseArray.size(); i++ )
			P = P.product( PoseArray[ i ] );
		m_Destination.push_back( P );

      // determine next CurTime by searching for the closest event with a timestamp > CurTime
      CurTime = DWARF::Util::maxTime;
      for ( unsigned i = 0; i < m_Sequences.size(); i++ )
         if ( Its[ i ] != m_Sequences[ i ]->end() && Its[ i ]->timestamp < CurTime )
            CurTime = Its[ i ]->timestamp;
      }

   // unlock sequences
   m_Destination.GetMutex().unlock();
   for ( unsigned i = 0; i < m_Sequences.size(); i++ )
      m_Sequences[ i ]->GetMutex().unlock();

	// send event only when finished completely
	if ( CurTime == DWARF::Util::maxTime )
		{
		DEBUGSTREAM( 20, "InfThread finished normally" );
		QApplication::postEvent( m_pReceiver, m_pEvent );
		}
	else
		{
		DEBUGSTREAM( 20, "InfThread stopped unfinished" );
		delete m_pEvent;
		}
	}

