/* ========================================================================
 * Copyright (C) 2000-2004  Technische Universitaet Muenchen
 * License: GPL v2 (or later)
 *
 * For further information please contact the DWARF team at
 * <dwarf-gnu@augmentedreality.de>
 *
 * MainWindowImpl.h
 * Daniel Pustka <pustka@in.tum.de> May 2004
 * Distributed Wearable Augmented Reality Framework - www.augmentedreality.de
 * For documentation and instructions, see the DWARF documentation website at
 * http://wwwbruegge.in.tum.de/projects/dwarf/doc/
 *
 * For further questions about DWARF, please contact us at
 * <dwarf@augmentedreality.de>
 * For questions about this particular service, contact
 * Daniel Pustka <pustka@in.tum.de>
 *
 * $Id: MainWindowImpl.cpp,v 1.2 2004/08/14 23:32:39 pustka Exp $
 */

#include <algorithm>
#include <qfiledialog.h>
#include <qlistbox.h>
#include <qsettings.h>

#include "MainWindowImpl.h"
#include "DisplayWindowImpl.h"
#include "PoseSequence.h"
#include "EkfWindowImpl.h"
#include "SigmaScaleWindowImpl.h"
#include "InfWindowImpl.h"
#include "TShiftWindowImpl.h"
#include "DerivWindowImpl.h"
#include "InvertWindowImpl.h"


MainWindowImpl::MainWindowImpl( QWidget* pParent, const char* pszName )
   : MainWindow( pParent, pszName )
   , m_bUpdating( false )
   {
   connect( g_pPoseSequenceCollection, SIGNAL(Changed()), this, SLOT( UpdateList() ) );
   QSettings Settings;
   m_sLastDir = Settings.readEntry( "/PoseTool/LastDirectory" );
   }


MainWindowImpl::~MainWindowImpl()
   {
   QSettings Settings;
   Settings.writeEntry( "/PoseTool/LastDirectory", m_sLastDir );
   }


void MainWindowImpl::Load()
   {
   // ask user for filename
   QString sFilename = QFileDialog::getOpenFileName( m_sLastDir, QString::null, this );
   if ( !sFilename.isNull() )
      {
      // keep last path
      int nPathLen = sFilename.findRev( '/' );
      m_sLastDir = sFilename.left( std::max( 0, nPathLen ) );

      // make display name
      std::string sName = g_pPoseSequenceCollection->CreateNewName(
         sFilename.right( sFilename.length() - nPathLen - 1 ) );

      // load
      PoseSequence* Sequence = g_pPoseSequenceCollection->AddSequence( sName.c_str() );
      Sequence->Load( sFilename );
      }
   }


void MainWindowImpl::Save()
   {
   // create default file name
   QString sItem = ListSequences->currentText();
   QString sFilename = m_sLastDir;
   if ( !sFilename.isEmpty() )
      sFilename += "/";
   sFilename += sItem;

   // ask user for filename
   sFilename = QFileDialog::getSaveFileName( sFilename, QString::null, this );

   // save
   if ( !sFilename.isNull() )
      g_pPoseSequenceCollection->GetSequence( sItem )->Save( sFilename );
   }


void MainWindowImpl::Record()
   {
   }


void MainWindowImpl::Play()
   {
   }


void MainWindowImpl::Remove()
   {
   // stop listbox from updating
   m_bUpdating = true;

   // remove selected
   for ( unsigned i = 0; i < ListSequences->count(); i++ )
      if ( ListSequences->isSelected( i ) )
         g_pPoseSequenceCollection->RemoveSequence( ListSequences->text( i ) );

   // update listbox
   m_bUpdating = false;
   UpdateList();
   }


void MainWindowImpl::Display()
   {
   PoseSequenceSubCollection PSSC;
   for ( unsigned i = 0; i < ListSequences->count(); i++ )
      if ( ListSequences->isSelected( i ) )
         PSSC.push_back( g_pPoseSequenceCollection->GetSequence( ListSequences->text( i ) ) );

   DisplayWindowImpl* pWnd = new DisplayWindowImpl();
   pWnd->SetSequences( PSSC );
   pWnd->show();
   }


void MainWindowImpl::KalmanFilter()
   {
   PoseSequenceSubCollection PSSC;
   for ( unsigned i = 0; i < ListSequences->count(); i++ )
      if ( ListSequences->isSelected( i ) )
         PSSC.push_back( g_pPoseSequenceCollection->GetSequence( ListSequences->text( i ) ) );

   EkfWindowImpl* pWnd = new EkfWindowImpl();
   pWnd->SetSequences( PSSC );
   pWnd->show();
   }


void MainWindowImpl::Difference()
   {
   }


void MainWindowImpl::UpdateList()
   {
   // don't update when a bigger change is in progress
   if ( m_bUpdating )
      return;

   // empty listbox
   ListSequences->clear();

   // insert items
   QMutexLocker l( &g_pPoseSequenceCollection->GetMutex() );

   const PoseSequenceCollection::CollectionType& Coll = g_pPoseSequenceCollection->GetCollection();
   PoseSequenceCollection::CollectionType::const_iterator it;
   for ( it = Coll.begin(); it != Coll.end(); it++ )
      ListSequences->insertItem( (*it)->GetName().c_str() );
   }


void MainWindowImpl::SigmaScale()
   {
   PoseSequenceSubCollection PSSC;
   for ( unsigned i = 0; i < ListSequences->count(); i++ )
      if ( ListSequences->isSelected( i ) )
         PSSC.push_back( g_pPoseSequenceCollection->GetSequence( ListSequences->text( i ) ) );

   SigmaScaleWindowImpl* pWnd = new SigmaScaleWindowImpl();
   pWnd->SetSequences( PSSC );
   pWnd->show();
   }


void MainWindowImpl::Infere()
   {
   PoseSequenceSubCollection PSSC;
   for ( unsigned i = 0; i < ListSequences->count(); i++ )
      if ( ListSequences->isSelected( i ) )
         PSSC.push_back( g_pPoseSequenceCollection->GetSequence( ListSequences->text( i ) ) );

   InfWindowImpl* pWnd = new InfWindowImpl();
   pWnd->SetSequences( PSSC );
   pWnd->show();
   }


void MainWindowImpl::TimeShift()
   {
   PoseSequenceSubCollection PSSC;
   for ( unsigned i = 0; i < ListSequences->count(); i++ )
      if ( ListSequences->isSelected( i ) )
         PSSC.push_back( g_pPoseSequenceCollection->GetSequence( ListSequences->text( i ) ) );

   TShiftWindowImpl* pWnd = new TShiftWindowImpl();
   pWnd->SetSequences( PSSC );
   pWnd->show();
   }


void MainWindowImpl::Differentiate()
   {
   PoseSequenceSubCollection PSSC;
   for ( unsigned i = 0; i < ListSequences->count(); i++ )
      if ( ListSequences->isSelected( i ) )
         PSSC.push_back( g_pPoseSequenceCollection->GetSequence( ListSequences->text( i ) ) );

   DerivWindowImpl* pWnd = new DerivWindowImpl();
   pWnd->SetSequences( PSSC );
   pWnd->show();
   }


void MainWindowImpl::Invert()
   {
   PoseSequenceSubCollection PSSC;
   for ( unsigned i = 0; i < ListSequences->count(); i++ )
      if ( ListSequences->isSelected( i ) )
         PSSC.push_back( g_pPoseSequenceCollection->GetSequence( ListSequences->text( i ) ) );

   InvertWindowImpl* pWnd = new InvertWindowImpl();
   pWnd->SetSequences( PSSC );
   pWnd->show();
   }

