/* ========================================================================
 * Copyright (C) 2000-2004  Technische Universitaet Muenchen
 * License: GPL v2 (or later)
 *
 * For further information please contact the DWARF team at
 * <dwarf-gnu@augmentedreality.de>
 *
 * PoseSequence.h
 * Daniel Pustka <pustka@in.tum.de> May 2004
 * Distributed Wearable Augmented Reality Framework - www.augmentedreality.de
 * For documentation and instructions, see the DWARF documentation website at
 * http://wwwbruegge.in.tum.de/projects/dwarf/doc/
 *
 * For further questions about DWARF, please contact us at
 * <dwarf@augmentedreality.de>
 * For questions about this particular service, contact
 * Daniel Pustka <pustka@in.tum.de>
 *
 * $Id: PoseSequence.cpp,v 1.2 2004/08/14 23:32:39 pustka Exp $
 */

#include <iostream>
#include <fstream>
#include <iomanip>
#include <stdio.h>
#include "PoseSequence.h"
#include "TimeUtil.h"
#include <debug.h>

PoseSequenceCollection* g_pPoseSequenceCollection = 0;


////////////////////////////////////////////////////////////////////////////
// operators for PoseData I/O

std::istream& operator>>( std::istream& is, DWARF::PoseData& Pose )
  {
  is >> Pose.position[0] >> Pose.position[1] >> Pose.position[2];
  is >> Pose.orientation[0] >> Pose.orientation[1] >> Pose.orientation[2] >> Pose.orientation[3];
  is >> Pose.timestamp.seconds;
  is.ignore() >> Pose.timestamp.microseconds;

  // additional fields added for the new ubitrack posedata
  is >> Pose.hasPosition >> Pose.hasOrientation >> Pose.confidence;
  for ( int i = 0; i < 6; i++ )
    for ( int j = 0; j < 6; j++ )
      is >> Pose.covariance[i][j];
  is >> Pose.timeError;

  DEBUGSTREAM( 50, "Loaded PoseData: t=" << Pose.timestamp.seconds << "." << Pose.timestamp.microseconds <<
   ", p=" << Pose.position[ 0 ] << ", " << Pose.position[ 1 ] << ", ..." );

  return is;
  }


std::ostream& operator<<( std::ostream& os, const DWARF::PoseData& pose )
   {
   os << pose.position[0] << " " << pose.position[1] << " " << pose.position[2] << " ";
   os << pose.orientation[0] << " " << pose.orientation[1] << " " << pose.orientation[2] << " " << pose.orientation[3] << " ";
   os << pose.timestamp.seconds << "." << std::setw( 6 ) << std::setfill( '0' ) << pose.timestamp.microseconds << " ";

  // additional fields added for the new ubitrack posedata
   os << pose.hasPosition << " " << pose.hasOrientation << " " << pose.confidence << " ";
   for ( int i = 0; i < 6; i++ )
      for ( int j = 0; j < 6; j++ )
         os << pose.covariance[i][j] << " ";
   os << pose.timeError;
   os << std::endl;

   return os;
   }


////////////////////////////////////////////////////////////////////////////
// PoseSequence

void PoseSequence::Load( const char* pszFilename )
   {
   std::ifstream is( pszFilename );
   if ( is.fail() )
      throw( std::string( "Could not open " ) + pszFilename );

   {
   QMutexLocker l( &m_Mutex );

   DWARF::PoseData PD;
   is >> PD;

   while ( !is.fail() )
      {
      push_back( PD );
      is >> PD;
      }
   }

   emit Changed();
   }


void PoseSequence::Save( const char* pszFilename )
   {
   std::ofstream os( pszFilename );
   if ( os.fail() )
      throw( std::string( "Could not open " ) + pszFilename );

   QMutexLocker l( &m_Mutex );

   for ( SequenceType::iterator it = begin(); it != end(); it++ )
      os << *it;
   }


////////////////////////////////////////////////////////////////////////////
// PoseSequenceCollection

PoseSequence* PoseSequenceCollection::AddSequence( const char* pszName )
   {
   {
   QMutexLocker l( &m_Mutex );
   m_Collection.push_back( std::auto_ptr<PoseSequence>( new PoseSequence( pszName ) ) );
   m_Lookup[ pszName ] = m_Collection.back();
   }

   emit Changed();

   return m_Collection.back();
   }


void PoseSequenceCollection::RemoveSequence( const char* pszName )
   {
   {
   QMutexLocker l( &m_Mutex );
   if ( m_Lookup[ pszName ]->IsReferenced() )
      throw( std::string( pszName + std::string( " is in use and cannot be removed" ) ) );

   // remove from collection
   PoseSequence* pSeq = m_Lookup[ pszName ];
   for ( unsigned i = 0; i < m_Collection.size(); i++ )
      if ( m_Collection[ i ] == pSeq )
         {
         m_Collection.erase( i );
         break;
         }

   // remove from lookup
   m_Lookup.erase( pszName );
   }

   emit Changed();
   }


std::string PoseSequenceCollection::CreateNewName( const char* pszName )
   {
   QMutexLocker l( &m_Mutex );

   std::string sName( pszName );
   int i = 2;
   while ( m_Lookup.find( sName ) != m_Lookup.end() )
      {
      char buf[ 10 ];
      sprintf( buf, "%u", i++ );
      sName = pszName;
      sName += "-";
      sName += buf;
      }

   return sName;
   }


////////////////////////////////////////////////////////////////////////////
// PoseSequenceSubCollection

DWARF::Time PoseSequenceSubCollection::GetFirstTime() const
   {
   DWARF::Time T = { 0x7FFFFFFF, 0x7FFFFFFF };
   for ( unsigned i = 0; i < size(); i++ )
      {
      // compute first position
      QMutexLocker l( &(*this)[ i ]->GetMutex() );
      if ( !(*this)[ i ]->empty() )
         {
         const DWARF::Time ET = (*this)[ i ]->front().timestamp;
         if ( ET < T )
            T = ET;
         }
      }

   return T;
   }


DWARF::Time PoseSequenceSubCollection::GetLastTime() const
   {
   DWARF::Time T = { 0, 0 };
   for ( unsigned i = 0; i < size(); i++ )
      {
      // compute first position
      QMutexLocker l( &(*this)[ i ]->GetMutex() );
      if ( !(*this)[ i ]->GetSequence().empty() )
         {
         const DWARF::Time ET = (*this)[ i ]->GetSequence().back().timestamp;
         if ( T < ET )
            T = ET;
         }
      }

   return T;
   }
