/* ========================================================================
 * Copyright (C) 2000-2004  Technische Universitaet Muenchen
 * License: GPL v2 (or later)
 *
 * For further information please contact the DWARF team at
 * <dwarf-gnu@augmentedreality.de>
 *
 * PoseSequence.h
 * Daniel Pustka <pustka@in.tum.de> May 2004
 * Distributed Wearable Augmented Reality Framework - www.augmentedreality.de
 * For documentation and instructions, see the DWARF documentation website at
 * http://wwwbruegge.in.tum.de/projects/dwarf/doc/
 *
 * For further questions about DWARF, please contact us at
 * <dwarf@augmentedreality.de>
 * For questions about this particular service, contact
 * Daniel Pustka <pustka@in.tum.de>
 *
 * $Id: PoseSequence.h,v 1.2 2004/08/14 23:32:39 pustka Exp $
 */

/**
 * @file PoseSequence.h
 * @brief contains PoseSequence and PoseSequenceCollection
 *
 * @author Daniel Pustka <pustka@in.tum.de>
 * @ingroup PoseTool
 */

#ifndef __POSESEQUENCE_H_
#define __POSESEQUENCE_H_

#include <list>
#include <map>
#include <vector>
#include <string>
#include <DWARF/PoseData.h>
#include <qobject.h>
#include <qmutex.h>
#include "auto_vector.h"


/**
 * contains a sequence of PoseData events
 */
class PoseSequence : public QObject, public std::list<DWARF::PoseData>
   {
   Q_OBJECT

   public:
      typedef std::list<DWARF::PoseData> SequenceType;

   protected:
      int m_nReferences;
      std::string m_sName;
      QMutex m_Mutex;
		int m_nPosDerivation;
		int m_nOriDerivation;

   public:
      PoseSequence( const char* pszName = "" )
         : m_nReferences( 0 )
         , m_sName( pszName )
			, m_nPosDerivation( 0 )
			, m_nOriDerivation( 0 )
         {}

      const std::string& GetName() const
         { return m_sName;}

      SequenceType& GetSequence()
         { return *this; }

      QMutex& GetMutex()
         { return m_Mutex; }

		void SetDerivations( int Pos, int Ori )
			{ m_nPosDerivation = Pos; m_nOriDerivation = Ori; }

		int GetPosDerivation() const
			{ return m_nPosDerivation; }

		int GetOriDerivation() const
			{ return m_nOriDerivation; }

      void AddRef()
         { m_nReferences++; }

      void RemoveRef()
         { m_nReferences--; }

      bool IsReferenced() const
         { return m_nReferences > 0; }

      void Load( const char* pszFilename );
      void Save( const char* pszFilename );

      void SignalChanged()
         { Changed(); }

   signals:
      void Changed();
   };


/**
 * contains a set of PoseSequences
 */
class PoseSequenceCollection : public QObject
   {
   Q_OBJECT

   public:
      typedef auto_vector<PoseSequence> CollectionType;

   protected:
      CollectionType m_Collection;
      std::map< std::string, PoseSequence* > m_Lookup;
      QMutex m_Mutex;

   public:
      PoseSequenceCollection()
         {}

      const CollectionType& GetCollection() const
         { return m_Collection; }

      QMutex& GetMutex()
         { return m_Mutex; }

      PoseSequence* GetSequence( const char* pszName )
         { return m_Lookup[ pszName ]; }

      PoseSequence* AddSequence( const char* pszName );

      void RemoveSequence( const char* pszName );

      /** creates a name that doesn't yet exist */
      std::string CreateNewName( const char* pszName );

   signals:
      void Changed();
   };


/**
 * contains pointers to a subcollection of posesequences
 */
class PoseSequenceSubCollection : public std::vector<PoseSequence*>
   {
   public:
      DWARF::Time GetFirstTime() const;
      DWARF::Time GetLastTime() const;
   };


/** one instance for the whole program */
extern PoseSequenceCollection* g_pPoseSequenceCollection;


#endif //__MAINWINDOW_H_
