/* ========================================================================
 * Copyright (C) 2000-2004  Technische Universitaet Muenchen
 * License: GPL v2 (or later)
 *
 * For further information please contact the DWARF team at
 * <dwarf-gnu@augmentedreality.de>
 *
 * PoseSequenceView.h
 * Daniel Pustka <pustka@in.tum.de> May 2004
 * Distributed Wearable Augmented Reality Framework - www.augmentedreality.de
 * For documentation and instructions, see the DWARF documentation website at
 * http://wwwbruegge.in.tum.de/projects/dwarf/doc/
 *
 * For further questions about DWARF, please contact us at
 * <dwarf@augmentedreality.de>
 * For questions about this particular service, contact
 * Daniel Pustka <pustka@in.tum.de>
 *
 * $Id: PoseSequenceView.cpp,v 1.2 2004/08/14 23:32:39 pustka Exp $
 */

#include "PoseSequenceView.h"
#include <qpainter.h>
#include <qevent.h>
#include <TimeUtil.h>
#include <qcolor.h>
#include <algorithm>
#include <math.h>
#include <debug.h>

static struct { QColor Normal, Sigma; } g_SeqColors[] = {
   { Qt::red, Qt::darkRed },
   { Qt::green, Qt::darkGreen },
   { Qt::yellow, Qt::darkYellow },
   { Qt::cyan, Qt::darkCyan },
   { Qt::magenta, Qt::darkMagenta },
   { Qt::blue, Qt::darkBlue } };


PoseSequenceView::PoseSequenceView( QWidget* pParent, const char* pszName )
   : QScrollView( pParent, pszName )
   , m_fPixelsPerSecond( 100.0 )
   , m_nHeight( 100 )
   , m_fDuration( 10.0 )
   , m_fMinValue( 0.0 )
   , m_fMaxValue( 0.5 )
   , m_bDrawVariance( false )
   {
   m_bPosEnabled[ 0 ] = true;
   for ( int i = 1; i < 3; i++ )
      m_bPosEnabled[ i ] = false;
   for ( int i = 0; i < 4; i++ )
      m_bOriEnabled[ i ] = false;
   setMouseTracking( true );
   }


PoseSequenceView::~PoseSequenceView()
   {
   // unreference sequences
   SetSequences( PoseSequenceSubCollection() );
   }


void PoseSequenceView::UpdateMinMaxValues( bool bLocal )
   {
   double fMin = 10e9;
   double fMax = -10e9;
   double fStartTime = 0.0;
   double fEndTime = m_fDuration;

   if ( bLocal )
      {
      fStartTime = contentsX() / m_fPixelsPerSecond;
      fEndTime = ( contentsX() + visibleWidth() ) / m_fPixelsPerSecond;
      }

   for ( unsigned i = 0; i < m_Sequences.size(); i++ )
      {
      QMutexLocker l( &m_Sequences[ i ]->GetMutex() );
      const PoseSequence::SequenceType& Seq = m_Sequences[ i ]->GetSequence();
      for ( PoseSequence::SequenceType::const_iterator it = Seq.begin(); it != Seq.end(); it++ )
         {
         // check time
         DWARF::Time T = it->timestamp - m_FirstEventTime;
         double fT = T.seconds + T.microseconds / 1000000.0;
         if ( fT < fStartTime )
            continue;
         else if ( fT > fEndTime )
            break;

         for ( int i = 0; i < 3; i++ )
            if ( m_bPosEnabled[ i ] )
               {
               fMin = std::min( fMin, it->position[ i ] );
               fMax = std::max( fMax, it->position[ i ] );
               }
         }
      }

   m_fMinValue = fMin;
   m_fMaxValue = fMax;
   }


void PoseSequenceView::SetSequences( const PoseSequenceSubCollection& Seqs )
   {
   // disconnect and unreference old sequences
   for ( unsigned i = 0; i < m_Sequences.size(); i++ )
      {
      disconnect( m_Sequences[ i ], SIGNAL( Changed() ), this, SLOT( SequenceChanged() ) );
      m_Sequences[ i ]->RemoveRef();
      }

   // copy sequence pointers
   m_Sequences = Seqs;

   if ( m_Sequences.empty() )
      return;

   // disconnect and reference sequences
   for ( unsigned i = 0; i < m_Sequences.size(); i++ )
      {
      connect( m_Sequences[ i ], SIGNAL( Changed() ), this, SLOT( SequenceChanged() ) );
      m_Sequences[ i ]->AddRef();
      }

   // compute first event and duration
   m_FirstEventTime = m_Sequences.GetFirstTime();
   DWARF::Time Duration = m_Sequences.GetLastTime() - m_FirstEventTime;
   m_fDuration = Duration.seconds + Duration.microseconds / 1000000.0;

   UpdateMinMaxValues();
   }


void PoseSequenceView::viewportResizeEvent( QResizeEvent* pEvent )
   {
   if ( pEvent->size().height() == m_nHeight )
      return;

   m_nHeight = pEvent->size().height();
   resizeContents( static_cast<int>( m_fPixelsPerSecond * m_fDuration ), m_nHeight );
   updateContents();
   }


void PoseSequenceView::DrawPosition( QPainter* pPainter, const PoseSequence::SequenceType& Seq,
   int nClipX1, int nClipX2, int iPos, int iColor )
   {
   if ( m_bDrawVariance )
      {
      pPainter->setPen( g_SeqColors[ iColor ].Sigma );
      QPoint OldPTop( 0, 0 );
      QPoint OldPBot( 0, 0 );
      for ( PoseSequence::SequenceType::const_iterator it = Seq.begin();
         it != Seq.end() && OldPTop.x() < nClipX2; it++ )
         {
         int nX = ToX( it->timestamp );
         double fSigma = sqrt( it->covariance[ iPos ][ iPos ] );
         QPoint PTop( nX, ToY( it->position[ iPos ] + fSigma ) );
         QPoint PBot( nX, ToY( it->position[ iPos ] - fSigma ) );
         if ( it != Seq.begin() && nX >= nClipX1 )
            {
            pPainter->drawLine( OldPTop, PTop );
            pPainter->drawLine( OldPBot, PBot );
            }
         OldPTop = PTop;
         OldPBot = PBot;
         }
      }

   pPainter->setPen( g_SeqColors[ iColor ].Normal );
   QPoint OldP( 0, 0 );
   for ( PoseSequence::SequenceType::const_iterator it = Seq.begin();
      it != Seq.end() && OldP.x() < nClipX2; it++ )
      {
      QPoint P( ToX( it->timestamp ), ToY( it->position[ iPos ] ) );
      if ( it != Seq.begin() && P.x() >= nClipX1 )
         {
         if ( m_bDrawDots )
            pPainter->drawEllipse( P.x() - 1, P.y() - 1, 3, 3 );
         pPainter->drawLine( OldP, P );
         }
      OldP = P;
      }
   }


void PoseSequenceView::DrawOrientation( QPainter* pPainter, const PoseSequence::SequenceType& Seq,
   int nClipX1, int nClipX2, int iOri, int iColor )
   {
/*   if ( m_bDrawVariance )
      {
      pPainter->setPen( g_SeqColors[ iColor ].Sigma );
      QPoint OldPTop( 0, 0 );
      QPoint OldPBot( 0, 0 );
      for ( PoseSequence::SequenceType::const_iterator it = Seq.begin();
         it != Seq.end() && OldPTop.x() < nClipX2; it++ )
         {
         int nX = ToX( it->timestamp );
         double fSigma = sqrt( it->covariance[ iOri ][ iOri ] );
         QPoint PTop( nX, ToY( it->position[ iPos ] + fSigma ) );
         QPoint PBot( nX, ToY( it->position[ iPos ] - fSigma ) );
         if ( it != Seq.begin() && nX >= nClipX1 )
            {
            pPainter->drawLine( OldPTop, PTop );
            pPainter->drawLine( OldPBot, PBot );
            }
         OldPTop = PTop;
         OldPBot = PBot;
         }
      } */

   pPainter->setPen( g_SeqColors[ iColor ].Normal );
   QPoint OldP( 0, 0 );
   for ( PoseSequence::SequenceType::const_iterator it = Seq.begin();
      it != Seq.end() && OldP.x() < nClipX2; it++ )
      {
      QPoint P( ToX( it->timestamp ), QToY( it->orientation[ iOri ] ) );
      if ( it != Seq.begin() && P.x() >= nClipX1 )
         {
         if ( m_bDrawDots )
            pPainter->drawEllipse( P.x() - 1, P.y() - 1, 3, 3 );
         pPainter->drawLine( OldP, P );
         }
      OldP = P;
      }
   }


void PoseSequenceView::drawContents( QPainter* pPainter, int nClipX, int nClipY, int nClipW, int nClipH )
   {
   DEBUGSTREAM( 20, "Redraw started" );
   int iColor = 0;
   for ( unsigned i = 0; i < m_Sequences.size(); i++ )
      {
      QMutexLocker l( &m_Sequences[ i ]->GetMutex() );
      const PoseSequence::SequenceType& Seq = m_Sequences[ i ]->GetSequence();
      if ( Seq.empty() )
         break;

      // draw x, y, z
      for ( int i = 0; i < 3; i++ )
         if ( m_bPosEnabled[ i ] )
            DrawPosition( pPainter, Seq, nClipX, nClipX + nClipW, i, iColor );

      // draw rx, ry, rz, rw
      for ( int i = 0; i < 4; i++ )
         if ( m_bOriEnabled[ i ] )
            DrawOrientation( pPainter, Seq, nClipX, nClipX + nClipW, i, iColor );

      iColor++;
      }
   DEBUGSTREAM( 20, "Redraw finished" );
   }


void PoseSequenceView::SequenceChanged()
   {
   updateContents();
   }


void PoseSequenceView::contentsMouseMoveEvent( QMouseEvent* e )
	{
	double fX = e->x() / m_fPixelsPerSecond + m_FirstEventTime.microseconds / 1000000.0 +
		m_FirstEventTime.seconds;
	double fY = ( m_nHeight - e->y() ) * ( m_fMaxValue - m_fMinValue ) / m_nHeight + m_fMinValue;
	CursorChanged( fX, fY );
	}


void PoseSequenceView::contentsMousePressEvent( QMouseEvent* e )
	{
	contentsMouseMoveEvent( e );
	}


void PoseSequenceView::contentsMouseReleaseEvent( QMouseEvent* e )
	{
	contentsMouseMoveEvent( e );
	}


void PoseSequenceView::enableX( bool b )
   {
   m_bPosEnabled[ 0 ] = b;
   updateContents();
   }


void PoseSequenceView::enableY( bool b )
   {
   m_bPosEnabled[ 1 ] = b;
   updateContents();
   }


void PoseSequenceView::enableZ( bool b )
   {
   m_bPosEnabled[ 2 ] = b;
   updateContents();
   }


void PoseSequenceView::enableRx( bool b )
   {
   m_bOriEnabled[ 0 ] = b;
   updateContents();
   }


void PoseSequenceView::enableRy( bool b )
   {
   m_bOriEnabled[ 1 ] = b;
   updateContents();
   }


void PoseSequenceView::enableRz( bool b )
   {
   m_bOriEnabled[ 2 ] = b;
   updateContents();
   }


void PoseSequenceView::enableRw( bool b )
   {
   m_bOriEnabled[ 3 ] = b;
   updateContents();
   }


void PoseSequenceView::enableVariance( bool b )
   {
   m_bDrawVariance = b;
   updateContents();
   }


void PoseSequenceView::enableDots( bool b )
   {
   m_bDrawDots = b;
   updateContents();
   }


void PoseSequenceView::ZoomYLocal()
   {
   UpdateMinMaxValues( true );
   updateContents();
   }


void PoseSequenceView::ZoomYGlobal()
   {
   UpdateMinMaxValues();
   updateContents();
   }


void PoseSequenceView::ZoomXIn()
   {
   m_fPixelsPerSecond *= 2;
   int nLeft = contentsX();
   resizeContents( static_cast<int>( m_fPixelsPerSecond * m_fDuration ), m_nHeight );
   setContentsPos( nLeft * 2, 0 );
   updateContents();
   }


void PoseSequenceView::ZoomXOut()
   {
   m_fPixelsPerSecond /= 2;
   int nLeft = contentsX();
   resizeContents( static_cast<int>( m_fPixelsPerSecond * m_fDuration ), m_nHeight );
   setContentsPos( nLeft / 2, 0 );
   updateContents();
   }


