/* ========================================================================
 * Copyright (C) 2000-2004  Technische Universitaet Muenchen
 * License: GPL v2 (or later)
 *
 * For further information please contact the DWARF team at
 * <dwarf-gnu@augmentedreality.de>
 *
 * SigmaScaleWindowImpl.h
 * Daniel Pustka <pustka@in.tum.de> May 2004
 * Distributed Wearable Augmented Reality Framework - www.augmentedreality.de
 * For documentation and instructions, see the DWARF documentation website at
 * http://wwwbruegge.in.tum.de/projects/dwarf/doc/
 *
 * For further questions about DWARF, please contact us at
 * <dwarf@augmentedreality.de>
 * For questions about this particular service, contact
 * Daniel Pustka <pustka@in.tum.de>
 *
 * $Id: SigmaScaleWindowImpl.cpp,v 1.2 2004/08/14 23:32:39 pustka Exp $
 */

#include "SigmaScaleWindowImpl.h"
#include <qslider.h>
#include <qcheckbox.h>
#include <qlineedit.h>
#include <qspinbox.h>
#include <qgroupbox.h>
#include <qmessagebox.h>
#include <debug.h>

static const double g_fLn10 = 2.30258509299404568401; // ln( 10 ) - used to compute 10^x from e^x


SigmaScaleWindowImpl::SigmaScaleWindowImpl( QWidget* pParent, const char* pszName )
   : SigmaScaleWindow( pParent, pszName )
   , m_pDestinationSequence( 0 )
   , m_fPosScale( 1.0 )
   , m_fOriScale( 1.0 )
   {
   setWFlags( Qt::WDestructiveClose );

   // init edit fields
   PosSliderChanged();
   OriSliderChanged();
   }


SigmaScaleWindowImpl::~SigmaScaleWindowImpl()
   {
   // unreference sequences
   SetSequences( PoseSequenceSubCollection() );
   }


void SigmaScaleWindowImpl::SetSequences( const PoseSequenceSubCollection& Seqs )
   {
   // disconnect and unreference old sequences
   for ( unsigned i = 0; i < m_Sequences.size(); i++ )
      {
      disconnect( m_Sequences[ i ], SIGNAL( Changed() ), this, SLOT( SequenceChanged() ) );
      m_Sequences[ i ]->RemoveRef();
      }

   // copy sequence pointers
   m_Sequences = Seqs;

   if ( m_Sequences.empty() )
      return;

   // disconnect and reference sequences
   for ( unsigned i = 0; i < m_Sequences.size(); i++ )
      {
      connect( m_Sequences[ i ], SIGNAL( Changed() ), this, SLOT( SequenceChanged() ) );
      m_Sequences[ i ]->AddRef();
      }

   // create a new sequence for the result
   std::string sName = g_pPoseSequenceCollection->CreateNewName(
      ( "SigScale-" + m_Sequences[ 0 ]->GetName() ).c_str() );
   m_pDestinationSequence = g_pPoseSequenceCollection->AddSequence( sName.c_str() );

   Recompute();
   }


void SigmaScaleWindowImpl::PosSliderChanged()
   {
   // update position
   m_fPosScale = exp( SliderPosScale->value() * g_fLn10 / 100 );
   char buf[10];
   sprintf( buf, "%1.1e", m_fPosScale );
   EditPosScale->setText( buf );

   Recompute();
   }


void SigmaScaleWindowImpl::OriSliderChanged()
   {
   // update position
   m_fOriScale = exp( SliderOriScale->value() * g_fLn10 / 100 );
   char buf[10];
   sprintf( buf, "%1.1e", m_fOriScale );
   EditOriScale->setText( buf );

   Recompute();
   }


void SigmaScaleWindowImpl::SequenceChanged()
   {
   Recompute();
   }


void SigmaScaleWindowImpl::Recompute()
   {
   if ( m_Sequences.empty() )
      return;

   DEBUGSTREAM( 20, "Scaling started" );

   // lock sequences
   {
   QMutexLocker lS( &m_Sequences[ 0 ]->GetMutex() );
   QMutexLocker lD( &m_pDestinationSequence->GetMutex() );
   DEBUGSTREAM( 20, "Scaling locked" );

   // erase destination
   m_pDestinationSequence->clear();
   m_pDestinationSequence->SetDerivations( m_Sequences[ 0 ]->GetPosDerivation(),
      m_Sequences[ 0 ]->GetOriDerivation() );

   PoseSequence::SequenceType::const_iterator it;
   for ( it = m_Sequences[ 0 ]->begin(); it != m_Sequences[ 0 ]->end(); it++ )
      {
      DWARF::PoseData P = *it;

      for ( int i = 0; i < 3; i++ )
         for ( int j = 0; j < 6; j++ )
            P.covariance[ i ][ j ] *= m_fPosScale;
      for ( int i = 0; i < 3; i++ )
         for ( int j = 0; j < 6; j++ )
            P.covariance[ j ][ i ] *= m_fPosScale;
      for ( int i = 3; i < 6; i++ )
         for ( int j = 0; j < 6; j++ )
            P.covariance[ i ][ j ] *= m_fOriScale;
      for ( int i = 3; i < 6; i++ )
         for ( int j = 0; j < 6; j++ )
            P.covariance[ j ][ i ] *= m_fOriScale;

      m_pDestinationSequence->push_back( P );
      }
   }
   DEBUGSTREAM( 20, "Scaling finished" );

   m_pDestinationSequence->SignalChanged();
   }
